#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2012       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GdkPixbuf
from gi.repository import GObject

import cairo
import sys, os
import platform

try:
    import bsddb3 as bsddb  # ok, in try/except

    BSDDB_STR = str(bsddb.__version__) + " " + str(bsddb.db.version())
except:
    try:
        import berkeleydb as bsddb

        BSDDB_STR = str(bsddb.__version__) + " " + str(bsddb.db.version())
    except:
        BSDDB_STR = "not found"

try:
    import sqlite3

    sqlite3_version_str = sqlite3.sqlite_version
except:
    sqlite3_version_str = "not found"

try:
    import orjson

    try:
        orjson_str = orjson.__version__
    except:  # any failure to 'get' the version
        orjson_str = "unknown version"

except ImportError:
    orjson_str = "not found"

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import ICON, SPLASH, URL_BUGTRACKER
from gramps.gen.constfunc import get_env_var
from gramps.version import VERSION
from ..display import display_help, display_url
from ..managedwindow import ManagedWindow


# -------------------------------------------------------------------------
#
# ErrorReportAssistant
#
# -------------------------------------------------------------------------
class ErrorReportAssistant(ManagedWindow, Gtk.Assistant):
    """
    Give the user an opportunity to report an error on the Gramps bug
    reporting system.
    """

    def __init__(self, error_detail, rotate_handler, ownthread=False, parent=None):
        Gtk.Assistant.__init__(self)
        ManagedWindow.__init__(self, None, [], self.__class__, modal=True)
        if parent is not None:  # the next two will not be set in ManagedWindow
            self.parent_window = parent
            self.other_modal_window = parent
        self.set_window(self, None, None, isWindow=True)
        if self.parent_window is not None:
            self.setup_configs("interface.errorreportassistant", 780, 520)

        try:
            # did we get a handler wrapping the error detail?
            self._error_detail = error_detail.get_formatted_log()
        except AttributeError:
            self._error_detail = error_detail

        self._rotate_handler = rotate_handler

        self._sys_information_text_buffer = None
        self._user_information_text_buffer = None
        self._error_details_text_buffer = None
        self._final_report_text_buffer = None

        self.set_title(_("Error Report Assistant"))
        self.connect("close", self.do_close)
        self.connect("cancel", self.do_close)
        self.connect("prepare", self.prepare)

        # create the assistant pages
        self.create_page_intro()
        self.build_page1()
        self.build_page2()
        self.build_page3()
        self.build_page4()
        self.build_page5()
        self.create_page_summary()

        self.show()  # ManagedWindow

        self.ownthread = ownthread
        if self.ownthread:
            Gtk.main()

    def do_close(self, *obj):
        """
        Close the assistant.
        """
        if self.parent_window is not None:
            self._save_position(save_config=False)  # the next line saves it
            self._save_size()
        self.destroy()
        if self.ownthread:
            Gtk.main_quit()

    def prepare(self, assistant, page):
        """
        Prepare pages prior to display.
        """
        self.page4_update()
        self.set_page_complete(page, True)

    def _copy_to_clipboard(self, obj=None):
        """
        Copy the bug report to the clipboard.
        """
        clipboard = Gtk.Clipboard.get_for_display(
            Gdk.Display.get_default(), Gdk.SELECTION_CLIPBOARD
        )
        clipboard.set_text(
            self._final_report_text_buffer.get_text(
                self._final_report_text_buffer.get_start_iter(),
                self._final_report_text_buffer.get_end_iter(),
                True,
            ),
            -1,
        )

        clipboard = t = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
        clipboard.set_text(
            self._final_report_text_buffer.get_text(
                self._final_report_text_buffer.get_start_iter(),
                self._final_report_text_buffer.get_end_iter(),
                True,
            ),
            -1,
        )

    def _start_email_client(self, obj=None):
        """
        Start an email client to send the report.
        """
        display_url(
            "mailto:gramps-bugs@lists.sourceforge.net?subject="
            '"bug report"&body="%s"'
            % self._final_report_text_buffer.get_text(
                self._final_report_text_buffer.get_start_iter(),
                self._final_report_text_buffer.get_end_iter(),
                True,
            )
        )

    def _start_gramps_bts_in_browser(self, obj=None):
        """
        Start a web browser to report the bug.
        """
        display_url(URL_BUGTRACKER)

    def _get_sys_information(self):
        """
        Get relevant system information.
        """
        distribution = ""  # print nothing if there's nothing to print
        if hasattr(os, "uname"):
            distribution = "Distribution: %s\n" % os.uname()[2]

        sqlite = f"sqlite version: {sqlite3_version_str} \n"

        return (
            "Gramps version: %s \n"
            "Python version: %s \n"
            "BSDDB version: %s \n"
            "%s"
            "orjson version: %s\n"
            "LANG: %s\n"
            "OS: %s\n"
            "%s\n"
            "GTK version    : %s\n"
            "gobject version: %s\n"
            "cairo version  : %s"
            % (
                str(VERSION),
                platform.python_version(),
                BSDDB_STR,
                sqlite,
                orjson_str,
                get_env_var("LANG", ""),
                platform.system(),
                distribution,
                "%d.%d.%d"
                % (
                    Gtk.get_major_version(),
                    Gtk.get_minor_version(),
                    Gtk.get_micro_version(),
                ),
                "%d.%d.%d" % GObject.pygobject_version,
                cairo.version_info,
            )
        )

    def _reset_error_details(self, obj=None):
        """
        Reset the error details buffer to its original contents.
        """
        self._error_details_text_buffer.set_text(
            "\n".join(self._rotate_handler.get_formatted_log(self._error_detail))
        )

    def _clear_error_details(self, obj=None):
        """
        Clear the error details buffer.
        """
        self._error_details_text_buffer.delete(
            self._error_details_text_buffer.get_start_iter(),
            self._error_details_text_buffer.get_end_iter(),
        )

    def _reset_sys_information(self, obj=None):
        """
        Reset the system information buffer to its original contents.
        """
        self._sys_information_text_buffer.set_text(self._get_sys_information())

    def _clear_sys_information(self, obj=None):
        """
        Clear the system information buffer.
        """
        self._sys_information_text_buffer.delete(
            self._sys_information_text_buffer.get_start_iter(),
            self._sys_information_text_buffer.get_end_iter(),
        )

    def _clear_user_information(self, obj=None):
        """
        Clear the user information buffer.
        """
        self._user_information_text_buffer.delete(
            self._user_information_text_buffer.get_start_iter(),
            self._user_information_text_buffer.get_end_iter(),
        )

    def create_page_intro(self):
        """
        Create the introduction page.
        """
        label = Gtk.Label(label=self.get_intro_text())
        label.set_line_wrap(True)

        image = Gtk.Image()
        image.set_from_file(SPLASH)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.pack_start(image, False, False, 5)
        box.pack_start(label, False, False, 5)

        page = box

        page.show_all()
        self.append_page(page)
        self.set_page_title(page, _("Report a bug"))
        self.set_page_type(page, Gtk.AssistantPageType.INTRO)

    def get_intro_text(self):
        """
        Return the text of the introduction page.
        """
        return _(
            "This is the Bug Reporting Assistant. It will "
            "help you to make a bug report to the Gramps "
            "developers that will be as detailed as possible.\n\n"
            "The assistant will ask you a few questions and will "
            "gather some information about the error that has "
            "occurred and the operating environment. "
            "At the end of the assistant you will be asked to "
            "file a bug report on the Gramps bug tracking system. "
            "The assistant will place the bug report on the clip board so "
            "that you can paste it into the form on the bug tracking "
            "website and review exactly what information you want to "
            "include."
        )

    def build_page1(self):
        """
        Build the error details page.
        """
        label = Gtk.Label(
            label=_(
                "If you can see that there is any personal "
                "information included in the error please remove "
                "it."
            )
        )
        label.set_halign(Gtk.Align.START)
        label.set_margin_top(4)
        label.set_margin_bottom(4)
        label.set_line_wrap(True)

        swin = Gtk.ScrolledWindow()
        swin.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        textview = Gtk.TextView()

        self._error_details_text_buffer = textview.get_buffer()
        self._reset_error_details()

        swin.add(textview)

        sw_frame = Gtk.Frame()
        sw_frame.add(swin)

        reset = Gtk.Button(label="Reset")
        reset.connect("clicked", self._reset_error_details)
        clear = Gtk.Button(label="Clear")
        clear.connect("clicked", self._clear_error_details)

        button_box = Gtk.ButtonBox()
        button_box.set_border_width(6)
        button_box.set_spacing(6)
        button_box.set_layout(Gtk.ButtonBoxStyle.END)

        button_box.pack_end(reset, False, False, 0)
        button_box.pack_end(clear, False, False, 0)

        error_details_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        error_details_box.pack_start(label, False, False, 0)
        error_details_box.pack_start(sw_frame, True, True, 0)
        error_details_box.pack_start(button_box, False, False, 0)
        error_details_box.set_hexpand(True)
        error_details_box.set_vexpand(True)

        error_details_box.set_property("margin", 12)

        error_details_frame = Gtk.Frame()
        error_details_frame.set_border_width(3)
        error_details_frame.set_label("<b>%s</b>" % _("Error Details"))
        error_details_frame.get_label_widget().set_use_markup(True)

        error_details_frame.add(error_details_box)

        side_label = Gtk.Label(
            label=_(
                "This is the detailed Gramps error "
                "information, don't worry if you do not "
                "understand it. You will have the opportunity "
                "to add further detail about the error "
                "in the following pages of the assistant."
            )
        )

        side_label.set_line_wrap(True)

        image = Gtk.Image()
        image.set_from_file(ICON)

        heading = Gtk.Box()
        heading.pack_start(side_label, True, True, 5)
        heading.pack_start(image, False, False, 5)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.pack_start(heading, False, False, 5)
        box.pack_start(error_details_frame, True, True, 0)

        page = box

        page.show_all()
        self.append_page(page)
        self.set_page_title(page, _("Error Details"))
        self.set_page_type(page, Gtk.AssistantPageType.CONTENT)

    def build_page2(self):
        """
        Build the system information page.
        """
        label = Gtk.Label(
            label=_(
                "Please check the information below and "
                "correct anything that you know to be wrong or "
                "remove anything that you would rather not have "
                "included in the bug report."
            )
        )
        label.set_halign(Gtk.Align.START)
        label.set_margin_top(4)
        label.set_margin_bottom(4)
        label.set_line_wrap(True)

        swin = Gtk.ScrolledWindow()
        swin.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        textview = Gtk.TextView()

        self._sys_information_text_buffer = textview.get_buffer()
        self._reset_sys_information()

        swin.add(textview)

        sw_frame = Gtk.Frame()
        sw_frame.add(swin)

        reset = Gtk.Button(label="Reset")
        reset.connect("clicked", self._reset_sys_information)
        clear = Gtk.Button(label="Clear")
        clear.connect("clicked", self._clear_sys_information)

        button_box = Gtk.ButtonBox()
        button_box.set_border_width(6)
        button_box.set_spacing(6)
        button_box.set_layout(Gtk.ButtonBoxStyle.END)

        button_box.pack_end(reset, False, False, 0)
        button_box.pack_end(clear, False, False, 0)

        sys_information_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        sys_information_box.pack_start(label, False, False, 0)
        sys_information_box.pack_start(sw_frame, True, True, 0)
        sys_information_box.pack_start(button_box, False, False, 0)
        sys_information_box.set_hexpand(True)
        sys_information_box.set_vexpand(True)
        sys_information_box.set_property("margin", 12)

        sys_information_frame = Gtk.Frame()
        sys_information_frame.set_border_width(3)
        sys_information_frame.set_label("<b>%s</b>" % _("System Information"))
        sys_information_frame.get_label_widget().set_use_markup(True)

        sys_information_frame.add(sys_information_box)

        side_label = Gtk.Label(
            label=_(
                "This is the information about your "
                "system that will help the developers to fix "
                "the bug."
            )
        )

        side_label.set_line_wrap(True)

        image = Gtk.Image()
        image.set_from_file(ICON)

        heading = Gtk.Box()
        heading.pack_start(side_label, True, True, 5)
        heading.pack_start(image, False, False, 5)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.pack_start(heading, False, False, 5)
        box.pack_start(sys_information_frame, True, True, 0)

        page = box

        page.show_all()
        self.append_page(page)
        self.set_page_title(page, _("System Information"))
        self.set_page_type(page, Gtk.AssistantPageType.CONTENT)

    def build_page3(self):
        """
        Build the further information page.
        """
        label = Gtk.Label(
            label=_(
                "Please provide as much information as you "
                "can about what you were doing when the error "
                "occurred."
            )
        )
        label.set_halign(Gtk.Align.START)
        label.set_margin_top(4)
        label.set_margin_bottom(4)
        label.set_line_wrap(True)

        swin = Gtk.ScrolledWindow()
        swin.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        textview = Gtk.TextView()

        self._user_information_text_buffer = textview.get_buffer()

        swin.add(textview)

        sw_frame = Gtk.Frame()
        sw_frame.add(swin)

        clear = Gtk.Button(label="Clear")
        clear.connect("clicked", self._clear_user_information)

        button_box = Gtk.ButtonBox()
        button_box.set_border_width(6)
        button_box.set_spacing(6)
        button_box.set_layout(Gtk.ButtonBoxStyle.END)

        button_box.pack_end(clear, False, False, 0)

        user_information_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        user_information_box.pack_start(label, False, False, 0)
        user_information_box.pack_start(sw_frame, True, True, 0)
        user_information_box.pack_start(button_box, False, False, 0)
        user_information_box.set_hexpand(True)
        user_information_box.set_vexpand(True)
        user_information_box.set_property("margin", 12)

        user_information_frame = Gtk.Frame()
        user_information_frame.set_border_width(3)
        user_information_frame.set_label("<b>%s</b>" % _("Further Information"))
        user_information_frame.get_label_widget().set_use_markup(True)

        user_information_frame.add(user_information_box)

        side_label = Gtk.Label(
            label=_(
                "This is your opportunity to describe "
                "what you were doing when the error occurred."
            )
        )

        side_label.set_line_wrap(True)

        image = Gtk.Image()
        image.set_from_file(ICON)

        heading = Gtk.Box()
        heading.pack_start(side_label, True, True, 5)
        heading.pack_start(image, False, False, 5)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.pack_start(heading, False, False, 5)
        box.pack_start(user_information_frame, True, True, 0)

        page = box

        page.show_all()
        self.append_page(page)
        self.set_page_title(page, _("Further Information"))
        self.set_page_type(page, Gtk.AssistantPageType.CONTENT)

    def build_page4(self):
        """
        Build the bug report summary page.
        """
        label = Gtk.Label(
            label=_(
                "Please check that the information is "
                "correct, do not worry if you don't understand the "
                "detail of the error information. Just make sure "
                "that it does not contain anything that you do not "
                "want to be sent to the developers."
            )
        )
        label.set_halign(Gtk.Align.START)
        label.set_margin_top(4)
        label.set_margin_bottom(4)
        label.set_line_wrap(True)

        swin = Gtk.ScrolledWindow()
        swin.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        textview = Gtk.TextView()
        textview.set_editable(False)
        textview.set_cursor_visible(False)

        self._final_report_text_buffer = textview.get_buffer()

        swin.add(textview)

        sw_frame = Gtk.Frame()
        sw_frame.add(swin)

        summary_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        summary_box.pack_start(label, False, False, 0)
        summary_box.pack_start(sw_frame, True, True, 0)
        summary_box.set_hexpand(True)
        summary_box.set_vexpand(True)
        summary_box.set_property("margin", 12)

        summary_frame = Gtk.Frame()
        summary_frame.set_border_width(3)
        summary_frame.set_label("<b>%s</b>" % _("Bug Report Summary"))
        summary_frame.get_label_widget().set_use_markup(True)

        summary_frame.add(summary_box)

        side_label = Gtk.Label(
            label=_(
                "This is the completed bug report. "
                "The next page of the assistant will help you "
                "to file a bug on the Gramps bug tracking "
                "system website."
            )
        )

        side_label.set_line_wrap(True)

        image = Gtk.Image()
        image.set_from_file(ICON)

        heading = Gtk.Box()
        heading.pack_start(side_label, True, True, 5)
        heading.pack_start(image, False, False, 5)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.pack_start(heading, False, False, 5)
        box.pack_start(summary_frame, True, True, 0)

        page = box

        page.show_all()
        self.append_page(page)
        self.set_page_title(page, _("Bug Report Summary"))
        self.set_page_type(page, Gtk.AssistantPageType.CONTENT)

    def build_page5(self):
        """
        Build the send bug report page.
        """
        label = Gtk.Label(
            label="%s <i>%s</i>"
            % (
                _(
                    "Use the two buttons below to first copy the bug report to the "
                    "clipboard and then open a webbrowser to file a bug report at "
                ),
                URL_BUGTRACKER,
            )
        )
        label.set_halign(Gtk.Align.START)
        label.set_margin_top(4)
        label.set_margin_bottom(4)
        label.set_line_wrap(True)
        label.set_use_markup(True)

        url_label = Gtk.Label(
            label=_(
                "Use this button to start a web browser "
                "and file a bug report on the Gramps bug "
                "tracking system."
            )
        )
        url_label.set_halign(Gtk.Align.START)
        url_label.set_margin_top(4)
        url_label.set_margin_bottom(4)
        url_label.set_line_wrap(True)
        url_label.set_size_request(200, -1)

        url_button = Gtk.Button(label="File bug report")
        url_button.connect("clicked", self._start_gramps_bts_in_browser)
        url_button_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        url_button_vbox.pack_start(url_button, True, False, 0)

        url_box = Gtk.Box()
        url_box.pack_start(url_label, True, True, 0)
        url_box.pack_start(url_button_vbox, False, False, 0)
        url_box.set_hexpand(True)
        url_box.set_property("margin", 12)

        url_frame = Gtk.Frame()
        url_frame.add(url_box)

        clip_label = Gtk.Label(
            label=_(
                "Use this button "
                "to copy the bug report onto the clipboard. "
                "Then go to the bug tracking website by using "
                "the button below, paste the report and click "
                "submit report"
            )
        )
        clip_label.set_halign(Gtk.Align.START)
        clip_label.set_margin_top(4)
        clip_label.set_margin_bottom(4)
        clip_label.set_line_wrap(True)
        clip_label.set_size_request(200, -1)

        clip_button = Gtk.Button(label="Copy to clipboard")
        clip_button.connect("clicked", self._copy_to_clipboard)
        clip_button_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        clip_button_vbox.pack_start(clip_button, True, False, 0)

        clip_box = Gtk.Box()
        clip_box.pack_start(clip_label, True, True, 0)
        clip_box.pack_start(clip_button_vbox, False, False, 0)
        clip_box.set_hexpand(True)
        clip_box.set_property("margin", 12)

        clip_frame = Gtk.Frame()
        clip_frame.add(clip_box)

        inner_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        inner_box.pack_start(label, False, False, 0)
        inner_box.pack_start(clip_frame, False, False, 0)
        inner_box.pack_start(url_frame, False, False, 0)
        inner_box.set_property("margin", 12)

        outer_frame = Gtk.Frame()
        outer_frame.set_border_width(3)
        outer_frame.set_label("<b>%s</b>" % _("Send Bug Report"))
        outer_frame.get_label_widget().set_use_markup(True)

        outer_frame.add(inner_box)

        side_label = Gtk.Label(
            label=_(
                "This is the final step. Use the "
                "buttons on this page to start a web browser "
                "and file a bug report on the Gramps bug "
                "tracking system."
            )
        )

        side_label.set_line_wrap(True)

        image = Gtk.Image()
        image.set_from_file(ICON)

        heading = Gtk.Box()
        heading.pack_start(side_label, True, True, 5)
        heading.pack_start(image, False, False, 5)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.pack_start(heading, False, False, 5)
        box.pack_start(outer_frame, True, True, 0)

        page = box

        page.show_all()
        self.append_page(page)
        self.set_page_title(page, _("Send Bug Report"))
        self.set_page_type(page, Gtk.AssistantPageType.CONTENT)

    def create_page_summary(self):
        """
        Create the summary page.
        """
        text = _(
            "Gramps is an Open Source project. Its success "
            "depends on its users. User feedback is important. "
            "Thank you for taking the time to submit a bug report."
        )
        label = Gtk.Label(label=text)
        label.set_line_wrap(True)

        image = Gtk.Image()
        image.set_from_file(SPLASH)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.pack_start(image, False, False, 5)
        box.pack_start(label, False, False, 5)

        page = box

        page.show_all()
        self.append_page(page)
        self.set_page_title(page, _("Complete"))
        self.set_page_type(page, Gtk.AssistantPageType.SUMMARY)

    def page4_update(self):
        """
        Update the contents of page 4 with any changes made.
        """
        self._final_report_text_buffer.set_text(
            "User Information: \n"
            + "===================\n\n"
            + self._user_information_text_buffer.get_text(
                self._user_information_text_buffer.get_start_iter(),
                self._user_information_text_buffer.get_end_iter(),
                True,
            )
            + "\n\n\nError Details: \n"
            + "===================\n\n"
            + self._error_details_text_buffer.get_text(
                self._error_details_text_buffer.get_start_iter(),
                self._error_details_text_buffer.get_end_iter(),
                True,
            )
            + "\n\nSystem Information: \n"
            + "===================\n\n"
            + self._sys_information_text_buffer.get_text(
                self._sys_information_text_buffer.get_start_iter(),
                self._sys_information_text_buffer.get_end_iter(),
                True,
            )
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2007 Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Declare constants used by database modules
"""

# -------------------------------------------------------------------------
#
# constants
#
# -------------------------------------------------------------------------
__all__ = (
    "DBPAGE",
    "DBMODE",
    "DBCACHE",
    "DBLOCKS",
    "DBOBJECTS",
    "DBUNDO",
    "DBEXT",
    "DBMODE_R",
    "DBMODE_W",
    "DBUNDOFN",
    "DBLOCKFN",
    "DBRECOVFN",
    "BDBVERSFN",
    "DBLOGNAME",
    "SCHVERSFN",
    "PCKVERSFN",
    "DBBACKEND",
    "PERSON_KEY",
    "FAMILY_KEY",
    "SOURCE_KEY",
    "CITATION_KEY",
    "EVENT_KEY",
    "MEDIA_KEY",
    "PLACE_KEY",
    "REPOSITORY_KEY",
    "NOTE_KEY",
    "REFERENCE_KEY",
    "TAG_KEY",
    "TXNADD",
    "TXNUPD",
    "TXNDEL",
    "CLASS_TO_KEY_MAP",
    "KEY_TO_CLASS_MAP",
    "KEY_TO_NAME_MAP",
)

DBEXT = ".db"  # File extension to be used for database files
DBUNDOFN = "undo.db"  # File name of 'undo' database
DBLOCKFN = "lock"  # File name of lock file
DBRECOVFN = "need_recover"  # File name of recovery file
BDBVERSFN = "bdbversion.txt"  # File name of Berkeley DB version file
DBBACKEND = "database.txt"  # File name of Database backend file
SCHVERSFN = "schemaversion.txt"  # File name of schema version file
PCKVERSFN = "pickleupgrade.txt"  # Indicator that pickle has been upgrade t Python3
DBLOGNAME = ".Db"  # Name of logger
DBMODE_R = "r"  # Read-only access
DBMODE_W = "w"  # Full Read/Write access
DBPAGE = 16384  # Size of the pages used to hold items in the database
DBMODE = 0o666  # Unix mode for database creation
DBCACHE = 0x4000000  # Size of the shared memory buffer pool
DBLOCKS = 100000  # Maximum number of locks supported
DBOBJECTS = 100000  # Maximum number of simultaneously locked objects
DBUNDO = 1000  # Maximum size of undo buffer
ARRAYSIZE = 1000  # The arraysize for a SQL cursor

PERSON_KEY = 0
FAMILY_KEY = 1
SOURCE_KEY = 2
EVENT_KEY = 3
MEDIA_KEY = 4
PLACE_KEY = 5
REPOSITORY_KEY = 6
REFERENCE_KEY = 7
NOTE_KEY = 8
TAG_KEY = 9
CITATION_KEY = 10

TXNADD, TXNUPD, TXNDEL = 0, 1, 2

CLASS_TO_KEY_MAP = {
    "Person": PERSON_KEY,
    "Family": FAMILY_KEY,
    "Source": SOURCE_KEY,
    "Citation": CITATION_KEY,
    "Event": EVENT_KEY,
    "Media": MEDIA_KEY,
    "Place": PLACE_KEY,
    "Repository": REPOSITORY_KEY,
    "Note": NOTE_KEY,
    "Tag": TAG_KEY,
}

KEY_TO_CLASS_MAP = {
    PERSON_KEY: "Person",
    FAMILY_KEY: "Family",
    SOURCE_KEY: "Source",
    CITATION_KEY: "Citation",
    EVENT_KEY: "Event",
    MEDIA_KEY: "Media",
    PLACE_KEY: "Place",
    REPOSITORY_KEY: "Repository",
    NOTE_KEY: "Note",
    TAG_KEY: "Tag",
}

KEY_TO_NAME_MAP = {
    PERSON_KEY: "person",
    FAMILY_KEY: "family",
    EVENT_KEY: "event",
    SOURCE_KEY: "source",
    CITATION_KEY: "citation",
    PLACE_KEY: "place",
    MEDIA_KEY: "media",
    REPOSITORY_KEY: "repository",
    # REFERENCE_KEY: 'reference',
    NOTE_KEY: "note",
    TAG_KEY: "tag",
}



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017,2024  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Event Reference class for Gramps
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .attrbase import AttributeBase
from .citationbase import CitationBase
from .const import DIFFERENT, EQUAL, IDENTICAL
from .eventroletype import EventRoleType
from .notebase import NoteBase
from .privacybase import PrivacyBase
from .refbase import RefBase
from .secondaryobj import SecondaryObject

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# EventRef
#
# -------------------------------------------------------------------------
class EventRef(
    PrivacyBase, NoteBase, AttributeBase, RefBase, CitationBase, SecondaryObject
):
    """
    Event reference class.

    This class is for keeping information about how the person relates
    to the referenced event.
    """

    def __init__(self, source=None):
        """
        Create a new EventRef instance, copying from the source if present.
        """
        PrivacyBase.__init__(self, source)
        CitationBase.__init__(self, source)
        NoteBase.__init__(self, source)
        AttributeBase.__init__(self, source)
        RefBase.__init__(self, source)
        if source:
            self.__role = EventRoleType(source.role)
        else:
            self.__role = EventRoleType()

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            PrivacyBase.serialize(self),
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            AttributeBase.serialize(self),
            RefBase.serialize(self),
            self.__role.serialize(),
        )

    def get_object_state(self):
        """
        Get the current object state as a dictionary.

        We override this method to handle the `role` property.
        """
        attr_dict = super().get_object_state()
        attr_dict["role"] = self.__role
        return attr_dict

    def set_object_state(self, attr_dict):
        """
        Set the current object state using information provided in the given
        dictionary.

        We override this method to handle the `role` property.
        """
        self.__role = attr_dict.pop("role")
        super().set_object_state(attr_dict)

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        # pylint: disable=import-outside-toplevel
        from .attribute import Attribute

        return {
            "type": "object",
            "title": _("Event reference"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "private": {"type": "boolean", "title": _("Private")},
                "citation_list": {
                    "type": "array",
                    "title": _("Citations"),
                    "items": {"type": "string", "maxLength": 50},
                },
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "attribute_list": {
                    "type": "array",
                    "items": Attribute.get_schema(),
                    "title": _("Attributes"),
                },
                "ref": {"type": "string", "maxLength": 50, "title": _("Event")},
                "role": EventRoleType.get_schema(),
            },
        }

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (privacy, citation_list, note_list, attribute_list, ref, role) = data
        PrivacyBase.unserialize(self, privacy)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        AttributeBase.unserialize(self, attribute_list)
        RefBase.unserialize(self, ref)
        self.__role = EventRoleType()
        self.__role.unserialize(role)
        return self

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.__role.string]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return self.attribute_list

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        return self.attribute_list

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return self.attribute_list

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: Returns the list of (classname, handle) tuples for referenced
                  objects.
        :rtype: list
        """
        ret = (
            self.get_referenced_citation_handles() + self.get_referenced_note_handles()
        )
        if self.ref:
            ret += [("Event", self.ref)]
        return ret

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through their
        children, reference primary objects..

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return self.get_citation_child_list()

    def is_equivalent(self, other):
        """
        Return if this eventref is equivalent, that is agrees in handle and
        role, to other.

        :param other: The eventref to compare this one to.
        :type other: EventRef
        :returns: Constant indicating degree of equivalence.
        :rtype: int
        """
        if self.ref != other.ref or self.role != other.role:
            return DIFFERENT
        if self.is_equal(other):
            return IDENTICAL
        return EQUAL

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this eventref.

        Lost: hlink and role of acquisition.

        :param acquisition: The eventref to merge with the present eventref.
        :type acquisition: EventRef
        """
        self._merge_privacy(acquisition)
        self._merge_attribute_list(acquisition)
        self._merge_citation_list(acquisition)
        self._merge_note_list(acquisition)

    def get_role(self):
        """
        Return the tuple corresponding to the preset role.
        """
        return self.__role

    def set_role(self, role):
        """
        Set the role according to the given argument.
        """
        self.__role.set(role)

    role = property(get_role, set_role, None, "Returns or sets role property")



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016  Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Geography for one person
"""
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import operator

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GLib
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import EventRoleType, EventType
from gramps.gen.config import config
from gramps.gen.datehandler import displayer
from gramps.gen.display.name import displayer as _nd
from gramps.gen.display.place import displayer as _pd
from gramps.gen.utils.place import conv_lat_lon
from gramps.gui.views.bookmarks import PersonBookmarks
from gramps.plugins.lib.maps import constants
from gramps.plugins.lib.maps.geography import GeoGraphyView

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
_LOG = logging.getLogger("GeoGraphy.geoperson")
KEY_TAB = Gdk.KEY_Tab
_ = glocale.translation.gettext

_UI_DEF = [
    """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="action">win.HomePerson</attribute>
          <attribute name="label" translatable="yes">_Home</attribute>
        </item>
      </section>
      </placeholder>
""",
    """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.PrintView</attribute>
          <attribute name="label" translatable="yes">Print...</attribute>
        </item>
      </section>
""",
    """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
    % _("Organize Bookmarks"),  # Following are the Toolbar items
    """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-home</property>
        <property name="action-name">win.HomePerson</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the home person</property>
        <property name="label" translatable="yes">_Home</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
    """
    <placeholder id='BarCommonEdit'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">document-print</property>
        <property name="action-name">win.PrintView</property>
        <property name="tooltip_text" translatable="yes">"""
    """Print or save the Map</property>
        <property name="label" translatable="yes">reference _Family</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
    """,
]

# pylint: disable=no-member
# pylint: disable=maybe-no-member
# pylint: disable=unused-variable
# pylint: disable=unused-argument


# -------------------------------------------------------------------------
#
# GeoView
#
# -------------------------------------------------------------------------
class GeoPerson(GeoGraphyView):
    """
    The view used to render person map.
    """

    CONFIGSETTINGS = (
        ("geography.path", constants.GEOGRAPHY_PATH),
        ("geography.zoom", 10),
        ("geography.zoom_when_center", 12),
        ("geography.show_cross", True),
        ("geography.lock", False),
        ("geography.center-lat", 0.0),
        ("geography.center-lon", 0.0),
        ("geography.use-keypad", True),
        ("geography.personal-map", ""),
        # ('geography.gps_mode', GPS_DISABLED),
        # ('geography.gps_update_rate', float(1.0)),
        # ('geography.max_gps_zoom', 16),
        # ('geography.gps_increment', GPS_INCREMENT),
        ("geography.map_service", constants.OPENSTREETMAP),
        ("geography.max_places", 5000),
        # specific to geoperson :
        ("geography.steps", 20),
        ("geography.maximum_lon_lat", 30),
        ("geography.speed", 100),
    )

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        GeoGraphyView.__init__(
            self,
            _("Person places map"),
            pdata,
            dbstate,
            uistate,
            PersonBookmarks,
            nav_group,
        )
        self.dbstate = dbstate
        self.uistate = uistate
        self.place_list = []
        self.place_without_coordinates = []
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        self.minyear = 9999
        self.maxyear = 0
        self.nbplaces = 0
        self.nbmarkers = 0
        self.sort = []
        self.additional_uis.append(self.additional_ui())
        self.no_show_places_in_status_bar = False
        self.already_started = False
        self.large_move = False
        self.cal = None

    def get_title(self):
        """
        Used to set the titlebar in the configuration window.
        """
        return _("GeoPerson")

    def get_stock(self):
        """
        Returns the name of the stock icon to use for the display.
        This assumes that this icon has already been registered
        as a stock icon.
        """
        return "geo-show-person"

    def get_viewtype_stock(self):
        """Type of view in category"""
        return "geo-show-person"

    def additional_ui(self):
        """
        Specifies the UIManager XML code that defines the menus and buttons
        associated with the interface.
        """
        return _UI_DEF

    def navigation_type(self):
        """
        Indicates the navigation type. Navigation type can be the string
        name of any of the primary objects.
        """
        return "Person"

    def goto_handle(self, handle=None):
        """
        Rebuild the tree with the given person handle as the root.
        """
        if self.osm is None:
            return
        active = self.get_active()
        self._createmap(None)
        self.uistate.modify_statusbar(self.dbstate)

    def build_tree(self):
        """
        This is called by the parent class when the view becomes visible. Since
        all handling of visibility is now in rebuild_trees, see that for more
        information.
        """
        pass

    def animate(self, menu, marks, index, stepyear):
        """
        Create all movements for the people's event.
        Yes, you can see the person moving.
        """
        if self.stop:  # no more database. stop to work
            return
        if len(marks) == 0:
            self.already_started = False
            return False
        i = int(index)
        next_i = i + 1
        if next_i == len(marks):
            self.already_started = False
            return False
        startlat = float(marks[i][3])
        startlon = float(marks[i][4])
        heading = 1
        if index == 0 and stepyear == 0:
            self.remove_all_gps()
            self.large_move = False
            self.osm.gps_add(startlat, startlon, heading)
        endlat = float(marks[next_i][3])
        endlon = float(marks[next_i][4])
        max_lon_lat = float(self._config.get("geography.maximum_lon_lat")) / 10
        if stepyear < 9000:
            if (abs(float(endlat) - float(startlat)) > max_lon_lat) or (
                abs(float(endlon) - float(startlon)) > max_lon_lat
            ):
                self.large_move = True
                stepyear = 9000
            else:
                self.large_move = False
        # year format = YYYYMMDD ( for sort )
        startyear = str(marks[i][6])[0:4]
        endyear = str(marks[next_i][6])[0:4]
        endmov = str(marks[len(marks) - 1][6])[0:4]
        years = int(endyear) - int(startyear)
        if years < 1:
            years = 1
        if stepyear > 8999:
            latstep = (endlat - startlat) / self._config.get("geography.steps")
            lonstep = (endlon - startlon) / self._config.get("geography.steps")
            startlat += latstep * (stepyear - 8999)
            startlon += lonstep * (stepyear - 8999)
        else:
            latstep = (endlat - startlat) / years
            lonstep = (endlon - startlon) / years
            stepyear = 1 if stepyear < 1 else stepyear
            startlat += latstep * stepyear
            startlon += lonstep * stepyear
        self.osm.gps_add(startlat, startlon, heading)
        stepyear += 1
        difflat = round(
            (startlat - endlat) if startlat > endlat else (endlat - startlat), 8
        )
        difflon = round(
            (startlon - endlon) if startlon > endlon else (endlon - startlon), 8
        )
        if difflat == 0.0 and difflon == 0.0:
            i += 1
            self.large_move = False
            stepyear = 1
        # if geography.speed = 100 => 100ms => 1s per 10 years.
        # For a 100 years person, it takes 10 secondes.
        # if large_move, one step is the difflat or difflon / geography.steps
        # in this case, stepyear is >= 9000
        # large move means longitude or latitude differences greater than
        # geography.maximum_lon_lat degrees.
        GLib.timeout_add(
            int(self._config.get("geography.speed")),
            self.animate,
            menu,
            marks,
            i,
            stepyear,
        )
        return False

    def _createmap(self, active):
        """
        Create all markers for each people's event in the database which has
        a lat/lon.
        @param: active is mandatory but unused in this view. Fix : 10088
        """
        if self.osm is None:
            return
        dbstate = self.dbstate
        self.cal = config.get("preferences.calendar-format-report")
        self.place_list = []
        self.place_without_coordinates = []
        self.places_found = []
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        self.minyear = 9999
        self.maxyear = 0
        latitude = ""
        longitude = ""
        self.nbplaces = 0
        self.nbmarkers = 0
        self.message_layer.clear_messages()
        self.message_layer.set_font_attributes(None, None, None)
        self.kml_layer.clear()
        person_handle = self.uistate.get_active("Person")
        person = None
        if person_handle:
            person = dbstate.db.get_person_from_handle(person_handle)
        if person is not None:
            # For each event, if we have a place, set a marker.
            self.load_kml_files(person)
            self.message_layer.add_message(
                _("Person places for %s") % _nd.display(person)
            )
            for event_ref in person.get_event_ref_list():
                if not event_ref:
                    continue
                event = dbstate.db.get_event_from_handle(event_ref.ref)
                self.load_kml_files(event)
                role = event_ref.get_role()
                eyear = (
                    str(
                        "%04d"
                        % event.get_date_object().to_calendar(self.cal).get_year()
                    )
                    + str(
                        "%02d"
                        % event.get_date_object().to_calendar(self.cal).get_month()
                    )
                    + str(
                        "%02d" % event.get_date_object().to_calendar(self.cal).get_day()
                    )
                )
                place_handle = event.get_place_handle()
                if place_handle:
                    place = dbstate.db.get_place_from_handle(place_handle)
                    if place:
                        longitude = place.get_longitude()
                        latitude = place.get_latitude()
                        latitude, longitude = conv_lat_lon(latitude, longitude, "D.D8")
                        descr = _pd.display(dbstate.db, place)
                        evt = EventType(event.get_type())
                        descr1 = _("%(eventtype)s : %(name)s") % {
                            "eventtype": evt,
                            "name": _nd.display(person),
                        }
                        self.load_kml_files(place)
                        # place.get_longitude and place.get_latitude return
                        # one string. We have coordinates when the two values
                        # contains non null string.
                        if longitude and latitude:
                            self._append_to_places_list(
                                descr,
                                evt,
                                _nd.display(person),
                                latitude,
                                longitude,
                                descr1,
                                eyear,
                                event.get_type(),
                                person.gramps_id,
                                place.gramps_id,
                                event.gramps_id,
                                role,
                            )
                        else:
                            self._append_to_places_without_coord(place.gramps_id, descr)
            family_list = person.get_family_handle_list()
            for family_hdl in family_list:
                family = self.dbstate.db.get_family_from_handle(family_hdl)
                if family is not None:
                    fhandle = family_list[0]  # first is primary
                    fam = dbstate.db.get_family_from_handle(fhandle)
                    father = mother = None
                    handle = fam.get_father_handle()
                    if handle:
                        father = dbstate.db.get_person_from_handle(handle)
                    descr1 = " - "
                    if father:
                        descr1 = "%s - " % _nd.display(father)
                    handle = fam.get_mother_handle()
                    if handle:
                        mother = dbstate.db.get_person_from_handle(handle)
                    if mother:
                        descr1 = "%s%s" % (descr1, _nd.display(mother))
                    for event_ref in family.get_event_ref_list():
                        if event_ref:
                            event = dbstate.db.get_event_from_handle(event_ref.ref)
                            self.load_kml_files(event)
                            role = event_ref.get_role()
                            if event.get_place_handle():
                                place_handle = event.get_place_handle()
                                if place_handle:
                                    place = dbstate.db.get_place_from_handle(
                                        place_handle
                                    )
                                    if place:
                                        longitude = place.get_longitude()
                                        latitude = place.get_latitude()
                                        (latitude, longitude) = conv_lat_lon(
                                            latitude, longitude, "D.D8"
                                        )
                                        descr = _pd.display(dbstate.db, place)
                                        evt = EventType(event.get_type())
                                        eyear = (
                                            str(
                                                "%04d"
                                                % event.get_date_object()
                                                .to_calendar(self.cal)
                                                .get_year()
                                            )
                                            + str(
                                                "%02d"
                                                % event.get_date_object()
                                                .to_calendar(self.cal)
                                                .get_month()
                                            )
                                            + str(
                                                "%02d"
                                                % event.get_date_object()
                                                .to_calendar(self.cal)
                                                .get_day()
                                            )
                                        )
                                        self.load_kml_files(place)
                                        if longitude and latitude:
                                            self._append_to_places_list(
                                                descr,
                                                evt,
                                                _nd.display(person),
                                                latitude,
                                                longitude,
                                                descr1,
                                                eyear,
                                                event.get_type(),
                                                person.gramps_id,
                                                place.gramps_id,
                                                event.gramps_id,
                                                role,
                                            )
                                        else:
                                            self._append_to_places_without_coord(
                                                place.gramps_id, descr
                                            )

            self.sort = sorted(self.place_list, key=operator.itemgetter(6))
            self._create_markers()

    def bubble_message(self, event, lat, lon, marks):
        self.menu = Gtk.Menu()
        menu = self.menu
        message = ""
        oldplace = ""
        prevmark = None
        for mark in marks:
            if oldplace != "":
                add_item = Gtk.MenuItem(label=message)
                add_item.show()
                menu.append(add_item)
                self.itemoption = Gtk.Menu()
                itemoption = self.itemoption
                itemoption.show()
                message = ""
                add_item.set_submenu(itemoption)
                modify = Gtk.MenuItem(label=_("Edit Event"))
                modify.show()
                modify.connect("activate", self.edit_event, event, lat, lon, prevmark)
                itemoption.append(modify)
                center = Gtk.MenuItem(label=_("Center on this place"))
                center.show()
                center.connect("activate", self.center_here, event, lat, lon, prevmark)
                itemoption.append(center)
            if mark[0] != oldplace:
                if message != "":
                    add_item = Gtk.MenuItem()
                    add_item.show()
                    menu.append(add_item)
                    self.itemoption = Gtk.Menu()
                    itemoption = self.itemoption
                    itemoption.show()
                    message = ""
                    add_item.set_submenu(itemoption)
                    modify = Gtk.MenuItem(label=_("Edit Event"))
                    modify.show()
                    modify.connect("activate", self.edit_event, event, lat, lon, mark)
                    itemoption.append(modify)
                    center = Gtk.MenuItem(label=_("Center on this place"))
                    center.show()
                    center.connect("activate", self.center_here, event, lat, lon, mark)
                    itemoption.append(center)
                message = "%s :" % mark[0]
                self.add_place_bubble_message(
                    event, lat, lon, marks, menu, message, mark
                )
                oldplace = mark[0]
                message = ""
            evt = self.dbstate.db.get_event_from_gramps_id(mark[10])
            # format the date as described in preferences.
            date = displayer.display(evt.get_date_object())
            if date == "":
                date = _("Unknown")
            if mark[11] == EventRoleType.PRIMARY:
                message = "(%s) %s : %s" % (date, mark[2], mark[1])
            elif mark[11] == EventRoleType.FAMILY:
                (father_name, mother_name) = self._get_father_and_mother_name(evt)
                message = "(%s) %s : %s - %s" % (
                    date,
                    mark[7],
                    father_name,
                    mother_name,
                )
            else:
                descr = evt.get_description()
                if descr == "":
                    descr = _("No description")
                message = "(%s) %s => %s" % (date, mark[11], descr)
            prevmark = mark
        add_item = Gtk.MenuItem(label=message)
        add_item.show()
        menu.append(add_item)
        self.itemoption = Gtk.Menu()
        itemoption = self.itemoption
        itemoption.show()
        add_item.set_submenu(itemoption)
        modify = Gtk.MenuItem(label=_("Edit Event"))
        modify.show()
        modify.connect("activate", self.edit_event, event, lat, lon, prevmark)
        itemoption.append(modify)
        center = Gtk.MenuItem(label=_("Center on this place"))
        center.show()
        center.connect("activate", self.center_here, event, lat, lon, prevmark)
        itemoption.append(center)
        menu.show()
        menu.popup_at_pointer(event)
        return 1

    def add_specific_menu(self, menu, event, lat, lon):
        """
        Add specific entry to the navigation menu.
        """
        menu.append(Gtk.SeparatorMenuItem())
        add_item = Gtk.MenuItem(label=_("Animate"))
        add_item.connect("activate", self.animate, self.sort, 0, 0)
        menu.append(add_item)
        return

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (("Person Filter",), ())

    def specific_options(self, configdialog):
        """
        Add specific entry to the preference menu.
        Must be done in the associated view.
        """
        grid = Gtk.Grid()
        grid.set_border_width(12)
        grid.set_column_spacing(6)
        grid.set_row_spacing(6)
        configdialog.add_text(
            grid,
            _("Animation speed in milliseconds (big value means slower)"),
            1,
            line_wrap=False,
        )
        configdialog.add_slider(grid, "", 2, "geography.speed", (100, 1000))
        configdialog.add_text(
            grid,
            _("How many steps between two markers when we are on large move ?"),
            3,
            line_wrap=False,
        )
        configdialog.add_slider(grid, "", 4, "geography.steps", (10, 100))
        configdialog.add_text(
            grid,
            _(
                "The minimum latitude/longitude to select large move.\n"
                "The value is in tenth of degree."
            ),
            5,
            line_wrap=False,
        )
        configdialog.add_slider(grid, "", 6, "geography.maximum_lon_lat", (5, 50))
        return _("The animation parameters"), grid



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007  Donald N. Allingham, Martin Hawlisch
# Copyright (C) 2009 Douglas S. Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

## Based on the paper:
##   http://www.cs.utah.edu/~draperg/research/fanchart/draperg_FHT08.pdf
## and the applet:
##   http://www.cs.utah.edu/~draperg/research/fanchart/demo/

## Found by redwood:
## http://www.gramps-project.org/bugs/view.php?id=2611

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.widgets.fanchart import (
    FanChartWidget,
    FanChartGrampsGUI,
    FORM_HALFCIRCLE,
    BACKGROUND_SCHEME1,
)
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class FanChartGramplet(FanChartGrampsGUI, Gramplet):
    """
    The Gramplet code that realizes the FanChartWidget.
    """

    def __init__(self, gui, nav_group=0):
        Gramplet.__init__(self, gui, nav_group)
        FanChartGrampsGUI.__init__(self, self.on_childmenu_changed)
        self.maxgen = 6
        self.background = BACKGROUND_SCHEME1
        self.childring = True
        self.flipupsidedownname = True
        self.twolinename = True
        self.radialtext = True
        self.fonttype = "Sans"
        self.grad_start = "#0000FF"
        self.grad_end = "#FF0000"
        self.generic_filter = None
        self.alpha_filter = 0.2
        self.form = FORM_HALFCIRCLE
        self.showid = False
        self.set_fan(FanChartWidget(self.dbstate, self.uistate, self.on_popup))
        # Replace the standard textview with the fan chart widget:
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.fan)
        # Make sure it is visible:
        self.fan.show()

    def init(self):
        self.set_tooltip(
            _(
                "Click to expand/contract person\nRight-click for "
                "options\nClick and drag in open area to rotate"
            )
        )

    def active_changed(self, handle):
        """
        Method called when active person changes.
        """
        # Reset everything but rotation angle (leave it as is)
        self.update()

    def on_childmenu_changed(self, obj, person_handle):
        """Callback for the pulldown menu selection, changing to the person
        attached with menu item."""
        dummy_obj = obj
        self.set_active("Person", person_handle)
        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2004  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python Modules
#
# -------------------------------------------------------------------------
import os
import tempfile
import zipfile
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .tabbeddoc import *
from ...const import PROGRAM_NAME, VERSION
from ...errors import ReportError

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

_XMLNS = """\
xmlns:office="%(urn)soffice:1.0"
xmlns:style="%(urn)sstyle:1.0"
xmlns:text="%(urn)stext:1.0"
xmlns:table="%(urn)stable:1.0"
xmlns:draw="%(urn)sdrawing:1.0"
xmlns:fo="%(urn)sxsl-fo-compatible:1.0"
xmlns:xlink="http://www.w3.org/1999/xlink"
xmlns:dc="http://purl.org/dc/elements/1.1/"
xmlns:meta="%(urn)smeta:1.0"
xmlns:number="%(urn)sdatastyle:1.0"
xmlns:svg="%(urn)ssvg-compatible:1.0"
xmlns:chart="%(urn)schart:1.0"
xmlns:dr3d="%(urn)sdr3d:1.0"
xmlns:math="http://www.w3.org/1998/Math/MathML"
xmlns:form="%(urn)sform:1.0"
xmlns:script="%(urn)sscript:1.0"
xmlns:dom="http://www.w3.org/2001/xml-events"
xmlns:xforms="http://www.w3.org/2002/xforms"
xmlns:xsd="http://www.w3.org/2001/XMLSchema"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
""" % {
    "urn": "urn:oasis:names:tc:opendocument:xmlns:"
}

_DOC_FONTS = """\
<office:font-face-decls>
    <style:font-face style:name="Nimbus Sans L"
        svg:font-family="'Nimbus Sans L'"
        style:font-family-generic="swiss"
        style:font-pitch="variable"/>
    <style:font-face style:name="DejaVu Sans"
        svg:font-family="'DejaVu San'"
        style:font-family-generic="system"
        style:font-pitch="variable"/>
</office:font-face-decls>
"""

_DOC_STYLES = """\
<office:automatic-styles>
    <style:style style:name="co1"
        style:family="table-column">
        <style:table-column-properties
            fo:break-before="auto"
            style:column-width="3cm"/>
    </style:style>
    <style:style style:name="ro1"
        style:family="table-row">
        <style:table-row-properties
            style:row-height="0.189in"
            fo:break-before="auto"
            style:use-optimal-row-height="true"/>
    </style:style>
    <style:style style:name="ta1"
        style:family="table"
        style:master-page-name="Default">
        <style:table-properties
            table:display="true"
            style:writing-mode="lr-tb"/>
    </style:style>
</office:automatic-styles>
"""

_STYLES_FONTS = """\
<office:font-face-decls>
    <style:font-face style:name="Times New Roman"
        svg:font-family="'Times New Roman'"
        style:font-family-generic="roman"
        style:font-pitch="variable"/>
    <style:font-face style:name="Arial"
        svg:font-family="Arial"
        style:font-family-generic="swiss"
        style:font-pitch="variable"/>
</office:font-face-decls>
"""

_STYLES_STYLES = """\
<office:styles>
    <style:default-style style:family="table-cell">
        <style:table-cell-properties
            style:decimal-places="2"/>
        <style:paragraph-properties
            style:tab-stop-distance="0.2835inch"/>
        <style:text-properties
            style:font-name="Arial"/>
    </style:default-style>

    <style:style style:name="Default"
        style:family="table-cell"
        style:data-style-name="N0"/>

    <style:default-style style:family="graphic">
        <style:text-properties fo:color="#000000"
            fo:font-family="'Times New Roman'"
            style:font-style-name=""
            style:font-family-generic="roman"
            style:font-pitch="variable"
            fo:font-size="12pt"
            fo:language="none"
            fo:country="none"
            style:text-autospace="ideograph-alpha"
            style:punctuation-wrap="simple"
            style:line-break="strict"/>
    </style:default-style>
</office:styles>
"""
_STYLES_AUTOMATIC = """\
<office:automatic-styles>
    <style:page-layout style:name="pm1">
        <style:header-style>
            <style:header-footer-properties
                fo:min-height="0.2957inch"
                fo:margin-left="0inch"
                fo:margin-right="0inch"
                fo:margin-bottom="0.0984inch"/>
        </style:header-style>
        <style:footer-style>
            <style:header-footer-properties
                fo:min-height="0.2957inch"
                fo:margin-left="0inch"
                fo:margin-right="0inch"
                fo:margin-top="0.0984inch"/>
        </style:footer-style>
    </style:page-layout>
    <style:page-layout style:name="pm2">
        <style:header-style>
            <style:header-footer-properties
                fo:min-height="0.2957inch"
                fo:margin-left="0inch"
                fo:margin-right="0inch"
                fo:margin-bottom="0.0984inch"
                fo:border="0.0346inch solid #000000"
                fo:border-top="0.0346inch solid #000000"
                fo:border-bottom="0.0346inch solid #000000"
                fo:border-left="0.0346inch solid #000000"
                fo:border-right="0.0346inch solid #000000"
                fo:padding="0.0071inch"
                fo:padding-top="0.0071inch"
                fo:padding-bottom="0.0071inch"
                fo:padding-left="0.0071inch"
                fo:padding-right="0.0071inch"
                fo:background-color="#c0c0c0"/>
        </style:header-style>
        <style:footer-style>
            <style:header-footer-properties
                fo:min-height="0.2957inch"
                fo:margin-left="0inch"
                fo:margin-right="0inch"
                fo:margin-top="0.0984inch"
                fo:border="0.0346inch solid #000000"
                fo:border-top="0.0346inch solid #000000"
                fo:border-bottom="0.0346inch solid #000000"
                fo:border-left="0.0346inch solid #000000"
                fo:border-right="0.0346inch solid #000000"
                fo:padding="0.0071inch"
                fo:padding-top="0.0071inch"
                fo:padding-bottom="0.0071inch"
                fo:padding-left="0.0071inch"
                fo:padding-right="0.0071inch"
                fo:background-color="#c0c0c0"/>
        </style:footer-style>
    </style:page-layout>
</office:automatic-styles>
"""

_STYLES_MASTER = """\
<office:master-styles>
    <style:master-page style:name="Default"
        style:page-layout-name="pm1">
        <style:header>
            <text:p>
                <text:sheet-name>???</text:sheet-name>
            </text:p>
        </style:header>
        <style:footer>
            <text:p>Page
                <text:page-number>1</text:page-number>
            </text:p>
        </style:footer>
    </style:master-page>
    <style:master-page style:name="Report"
        style:page-layout-name="pm2">
        <style:header>
            <style:region-left>
                <text:p>
                    <text:sheet-name>???</text:sheet-name>
                    (<text:file-name>???</text:file-name>)
                </text:p>
            </style:region-left>
            <style:region-right>
                <text:p>
                    <text:date style:data-style-name="N2"
                        text:date-value="2001-05-16">05/16/2001
                    </text:date>,
                        <text:time>10:53:17</text:time>
                </text:p>
            </style:region-right>
        </style:header>
        <style:footer>
            <text:p>Page
                <text:page-number>1</text:page-number> /
                <text:page-count>99</text:page-count>
            </text:p>
        </style:footer>
    </style:master-page>
</office:master-styles>
"""

_MANIFEST = """\
<?xml version="1.0" encoding="UTF-8"?>
<manifest:manifest
    xmlns:manifest=
        "urn:oasis:names:tc:opendocument:xmlns:manifest:1.0">
    <manifest:file-entry
        manifest:media-type=
            "application/vnd.oasis.opendocument.spreadsheet"
        manifest:full-path="/"/>
    <manifest:file-entry manifest:media-type=""
        manifest:full-path="Pictures/"/>
    <manifest:file-entry manifest:media-type="text/xml"
        manifest:full-path="content.xml"/>
    <manifest:file-entry manifest:media-type="text/xml"
        manifest:full-path="styles.xml"/>
    <manifest:file-entry manifest:media-type="text/xml"
        manifest:full-path="meta.xml"/>
</manifest:manifest>
"""

_META = """\
<?xml version="1.0" encoding="UTF-8"?>
<office:document-meta
    xmlns:office=
        "urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:meta=
        "urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
    office:class="text" office:version="1.0">
    <office:meta>
        <meta:generator>
            %(program)s %(version)s
        </meta:generator>
        <meta:initial-creator>
            %(name)s
        </meta:initial-creator>
        <meta:creation-date>
            %(time)s
        </meta:creation-date>
        <dc:creator>
            %(name)s
        </dc:creator>
        <dc:date>
            %(time)s
        </dc:date>
        <meta:print-date>
            0-00-00T00:00:00
        </meta:print-date>
        <dc:language>en-US</dc:language>
        <meta:editing-cycles>1</meta:editing-cycles>
        <meta:editing-duration>PT0S</meta:editing-duration>
        <meta:user-defined meta:name="Info 0"/>
        <meta:user-defined meta:name="Info 1"/>
        <meta:user-defined meta:name="Info 2"/>
        <meta:user-defined meta:name="Info 3"/>
    </office:meta>
</office:document-meta>
"""


# -------------------------------------------------------------------------
#
# ODSTab
#
# -------------------------------------------------------------------------
class ODSTab(TabbedDoc):
    def __init__(self, columns):
        TabbedDoc.__init__(self, columns)
        self.f = None
        self.filename = None
        self.level = 0
        self.time = "0000-00-00T00:00:00"

    def open(self, filename):
        import time

        t = time.localtime(time.time())
        self.time = "%04d-%02d-%02dT%02d:%02d:%02d" % t[:6]

        self.filename = filename
        if not filename.endswith(".ods"):
            self.filename += ".ods"

        try:
            self.content_xml = tempfile.mktemp()
            self.f = open(self.content_xml, "wb")
        except IOError as msg:
            raise ReportError(_("Could not create %s") % self.content_xml, msg)
        except:
            raise ReportError(_("Could not create %s") % self.content_xml)

        self.f = open(self.content_xml, "w", encoding="utf-8")
        self.f.write(
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            + "<office:document-content "
            + _XMLNS
            + 'office:version="1.0"> '
            + "<office:script/>\n"
        )

        self.f.write(_DOC_FONTS)
        self.f.write(_DOC_STYLES)
        self.f.write("<office:body>\n" "<office:spreadsheet>\n")

    def close(self):
        self.f.write(
            "</office:spreadsheet>\n" "</office:body>\n" "</office:document-content>\n"
        )
        self.f.close()
        self._write_styles_file()
        self._write_manifest()
        self._write_meta_file()
        self._write_mimetype_file()
        self._write_zip()

    def start_row(self):
        self.f.write('<table:table-row table:style-name="')
        self.f.write("ro1")
        self.f.write('">\n')

    def end_row(self):
        self.f.write("</table:table-row>\n")

    def write_cell(self, text):
        self.f.write('<table:table-cell office:value-type="string">')
        self.f.write(">\n")

        self.f.write("<text:p>")
        if text is not None:  # it must not be just 'if text'
            text = text.replace("&", "&amp;")  # Must be first
            text = text.replace("<", "&lt;")
            text = text.replace(">", "&gt;")
            text = text.replace("\t", "<text:tab-stop/>")
            text = text.replace("\n", "<text:line-break/>")
            self.f.write(str(text))

        self.f.write("</text:p>\n")
        self.f.write("</table:table-cell>\n")

    #        for col in range(1,self.span):
    #            self.f.write('<table:covered-table-cell/>\n')

    def _write_zip(self):
        try:
            file = zipfile.ZipFile(self.filename, "w", zipfile.ZIP_DEFLATED)
        except IOError as msg:
            errmsg = "%s\n%s" % (_("Could not create %s") % self.filename, msg)
            raise ReportError(errmsg)
        except:
            raise ReportError(_("Could not create %s") % self.filename)

        file.write(self.manifest_xml, str("META-INF/manifest.xml"))
        file.write(self.content_xml, str("content.xml"))
        file.write(self.meta_xml, str("meta.xml"))
        file.write(self.styles_xml, str("styles.xml"))
        file.write(self.mimetype, str("mimetype"))
        file.close()

        os.unlink(self.manifest_xml)
        os.unlink(self.content_xml)
        os.unlink(self.meta_xml)
        os.unlink(self.styles_xml)

    def _write_styles_file(self):
        self.styles_xml = tempfile.mktemp()

        try:
            self.f = open(self.styles_xml, "wb")
        except IOError as msg:
            errmsg = "%s\n%s" % (_("Could not create %s") % self.styles_xml, msg)
            raise ReportError(errmsg)
        except:
            raise ReportError(_("Could not create %s") % self.styles_xml)

        self.f = open(self.styles_xml, "w")
        self.f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        self.f.write("<office:document-styles " + _XMLNS + 'office:version="1.0"> ')
        self.f.write(_STYLES_FONTS)
        self.f.write(_STYLES_STYLES)
        self.f.write(_STYLES_AUTOMATIC)
        self.f.write(_STYLES_MASTER)

        self.f.write("</office:document-styles>\n")
        self.f.close()

    def start_page(self):
        self.f.write('<table:table table:name="ta1">')
        for col in range(0, self.columns):
            self.f.write('<table:table-column table:style-name="co1"/>\n')

    def end_page(self):
        self.f.write("</table:table>\n")

    def _write_manifest(self):
        self.manifest_xml = tempfile.mktemp()

        try:
            self.f = open(self.manifest_xml, "wb")
        except IOError as msg:
            errmsg = "%s\n%s" % (_("Could not create %s") % self.manifest_xml, msg)
            raise ReportError(errmsg)
        except:
            raise ReportError(_("Could not create %s") % self.manifest_xml)

        self.f = open(self.manifest_xml, "w")
        self.f.write(_MANIFEST)
        self.f.close()

    def _write_meta_file(self):
        self.meta_xml = tempfile.mktemp()

        try:
            self.f = open(self.meta_xml, "wb")
        except IOError as msg:
            errmsg = "%s\n%s" % (_("Could not create %s") % self.meta_xml, msg)
            raise ReportError(errmsg)
        except:
            raise ReportError(_("Could not create %s") % self.meta_xml)

        self.f = open(self.meta_xml, "w")

        self.f.write(
            _META
            % {
                "program": PROGRAM_NAME,
                "version": VERSION,
                "name": self.name,
                "time": self.time,
            }
        )
        self.f.close()

    def _write_mimetype_file(self):
        self.mimetype = tempfile.mktemp()

        try:
            self.f = open(self.mimetype, "wb")
        except IOError as msg:
            errmsg = "%s\n%s" % (_("Could not create %s") % self.mimetype, msg)
            raise ReportError(errmsg)
        except:
            raise ReportError(_("Could not create %s") % self.mimetype)

        self.f = open(self.mimetype, "w")
        self.f.write("application/vnd.oasis.opendocument.spreadsheet")
        self.f.close()


if __name__ == "__main__":
    file = ODSTab(3)
    file.open("test")
    file.start_page()
    for i in [("one", "two", "three"), ('fo"ur', "fi,ve", "six")]:
        file.start_row()
        for j in i:
            file.write_cell(j)
        file.end_row()
    file.end_page()
    file.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015 Douglas S. Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Base class for undo/redo functionality.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import time
from abc import ABCMeta, abstractmethod
from collections import deque


# -------------------------------------------------------------------------
#
# DbUndo class
#
# -------------------------------------------------------------------------
class DbUndo(metaclass=ABCMeta):
    """
    Base class for the Gramps undo/redo manager.  Needs to be subclassed
    for use with a real backend.
    """

    __slots__ = ("undodb", "db", "undo_history_timestamp", "undoq", "redoq")

    def __init__(self, db):
        """
        Class constructor. Set up main instance variables
        """
        self.db = db
        self.undoq = deque()
        self.redoq = deque()
        self.undo_history_timestamp = time.time()

    def clear(self):
        """
        Clear the undo/redo list (but not the backing storage)
        """
        self.undoq.clear()
        self.redoq.clear()
        self.undo_history_timestamp = time.time()

    def __enter__(self, value):
        """
        Context manager method to establish the context
        """
        self.open(value)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager method to finish the context
        """
        if exc_type is None:
            self.close()
        return exc_type is None

    @abstractmethod
    def open(self, value):
        """
        Open the backing storage.  Needs to be overridden in the derived
        class.
        """

    @abstractmethod
    def close(self):
        """
        Close the backing storage.  Needs to be overridden in the derived
        class.
        """

    @abstractmethod
    def append(self, value):
        """
        Add a new entry on the end.  Needs to be overridden in the derived
        class.
        """

    @abstractmethod
    def __getitem__(self, index):
        """
        Returns an entry by index number.  Needs to be overridden in the
        derived class.
        """

    @abstractmethod
    def __setitem__(self, index, value):
        """
        Set an entry to a value.  Needs to be overridden in the derived class.
        """

    @abstractmethod
    def __len__(self):
        """
        Returns the number of entries.  Needs to be overridden in the derived
        class.
        """

    @abstractmethod
    def _redo(self, update_history):
        """ """

    @abstractmethod
    def _undo(self, update_history):
        """ """

    def commit(self, txn, msg):
        """
        Commit the transaction to the undo/redo database.  "txn" should be
        an instance of Gramps transaction class
        """
        txn.set_description(msg)
        txn.timestamp = time.time()
        self.undoq.append(txn)
        self._after_commit(txn)

    def _after_commit(self, transaction):
        """
        Post-transaction commit processing.
        """

    def undo(self, update_history=True):
        """
        Undo a previously committed transaction
        """
        if self.db.readonly or self.undo_count == 0:
            return False
        return self._undo(update_history)

    def redo(self, update_history=True):
        """
        Redo a previously committed, then undone, transaction
        """
        if self.db.readonly or self.redo_count == 0:
            return False
        return self._redo(update_history)

    undo_count = property(lambda self: len(self.undoq))
    redo_count = property(lambda self: len(self.redoq))



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015-2016 Douglas S. Blank <doug.blank@gmail.com>
# Copyright (C) 2016-2017 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Backend for SQLite database.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging
import os
import re
import sqlite3

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.db.dbconst import ARRAYSIZE
from gramps.plugins.db.dbapi.dbapi import DBAPI

_ = glocale.translation.gettext

sqlite3.paramstyle = "qmark"


# -------------------------------------------------------------------------
#
# SQLite class
#
# -------------------------------------------------------------------------
class SQLite(DBAPI):
    """
    SQLite interface.
    """

    def get_summary(self):
        """
        Return a dictionary of information about this database backend.
        """
        summary = super().get_summary()
        summary.update(
            {
                _("Database version"): sqlite3.sqlite_version,
                _("Database module location"): sqlite3.__file__,
            }
        )
        return summary

    def _initialize(self, directory, username, password):
        if directory == ":memory:":
            path_to_db = ":memory:"
        else:
            path_to_db = os.path.join(directory, "sqlite.db")
        self.dbapi = Connection(path_to_db)


# -------------------------------------------------------------------------
#
# Connection class
#
# -------------------------------------------------------------------------
class Connection:
    """
    The Sqlite class is an interface between the DBAPI class which is the Gramps
    backend for the DBAPI interface and the sqlite3 python module.
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new Sqlite instance.

        This connects to a sqlite3 database and creates a cursor instance.

        :param args: arguments to be passed to the sqlite3 connect class at
                     creation.
        :type args: list
        :param kwargs: arguments to be passed to the sqlite3 connect class at
                       creation.
        :type kwargs: list
        """
        self.log = logging.getLogger(".sqlite")
        self.__connection = sqlite3.connect(*args, **kwargs)
        self.__cursor = self.__connection.cursor()
        self.__connection.create_function("regexp", 2, regexp)
        self.__collations = []
        self.__tmap = str.maketrans("-.@=;", "_____")
        self.check_collation(glocale)

    def check_collation(self, locale):
        """
        Checks that a collation exists and if not creates it.

        :param locale: Locale to be checked.
        :param type: A GrampsLocale object.
        """
        # PySQlite3 permits only ascii alphanumerics and underscores in
        # collation names so first translate any old-style Unicode locale
        # delimiters to underscores.
        collation = locale.get_collation().translate(self.__tmap)
        if collation not in self.__collations:
            self.__connection.create_collation(collation, locale.strcoll)
            self.__collations.append(collation)
        return collation

    def execute(self, *args, **kwargs):
        """
        Executes an SQL statement.

        :param args: arguments to be passed to the sqlite3 execute statement
        :type args: list
        :param kwargs: arguments to be passed to the sqlite3 execute statement
        :type kwargs: list
        """
        self.log.debug(args)
        self.__cursor.execute(*args, **kwargs)

    def fetchone(self):
        """
        Fetches the next row of a query result set, returning a single sequence,
        or None when no more data is available.
        """
        return self.__cursor.fetchone()

    def fetchall(self):
        """
        Fetches the next set of rows of a query result, returning a list. An
        empty list is returned when no more rows are available.
        """
        return self.__cursor.fetchall()

    def begin(self):
        """
        Start a transaction manually. This transactions usually persist until
        the next COMMIT or ROLLBACK command.
        """
        self.log.debug("BEGIN TRANSACTION;")
        self.execute("BEGIN TRANSACTION;")

    def commit(self):
        """
        Commit the current transaction.
        """
        self.log.debug("COMMIT;")
        self.__connection.commit()

    def rollback(self):
        """
        Roll back any changes to the database since the last call to commit().
        """
        self.log.debug("ROLLBACK;")
        self.__connection.rollback()

    def table_exists(self, table):
        """
        Test whether the specified SQL database table exists.

        :param table: table name to check.
        :type table: str
        :returns: True if the table exists, false otherwise.
        :rtype: bool
        """
        self.execute(
            "SELECT COUNT(*) "
            "FROM sqlite_master "
            f"WHERE type='table' AND name='{table}';"
        )
        return self.fetchone()[0] != 0

    def column_exists(self, table, column):
        """
        Test whether the specified SQL column exists in the specified table.

        :param table: table name to check.
        :type table: str
        :param column: column name to check.
        :type column: str
        :returns: True if the column exists, False otherwise.
        :rtype: bool
        """
        self.execute(
            "SELECT COUNT(*) "
            f"FROM pragma_table_info('{table}') "
            f"WHERE name = '{column}'"
        )
        return self.fetchone()[0] != 0

    def drop_column(self, table_name, column_name):
        # DROP COLUMN is available with Sqlite v 3.35.0, released 2021-03-12
        db_ver = sqlite3.sqlite_version.split(".")
        if int(db_ver[0]) == 3 and int(db_ver[1]) >= 35:
            self.execute(f"ALTER TABLE {table_name} DROP COLUMN {column_name};")

    def close(self):
        """
        Close the current database.
        """
        self.log.debug("closing database...")
        self.__connection.close()

    def cursor(self):
        """
        Return a new cursor.
        """
        return Cursor(self.__connection)


# -------------------------------------------------------------------------
#
# Cursor class
#
# -------------------------------------------------------------------------
class Cursor:
    """
    Exposes access to a SQLite cursor as an iterator
    """

    def __init__(self, connection):
        self.__connection = connection

    def __enter__(self):
        self.__cursor = self.__connection.cursor()
        self.__cursor.arraysize = ARRAYSIZE
        return self

    def __exit__(self, *args, **kwargs):
        self.__cursor.close()

    def execute(self, *args, **kwargs):
        """
        Executes an SQL statement.

        :param args: arguments to be passed to the sqlite3 execute statement
        :type args: list
        :param kwargs: arguments to be passed to the sqlite3 execute statement
        :type kwargs: list
        """
        self.__cursor.execute(*args, **kwargs)

    def fetchmany(self):
        """
        Fetches the next set of rows of a query result, returning a list. An
        empty list is returned when no more rows are available.
        """
        return self.__cursor.fetchmany()


def regexp(expr, value):
    """
    A user defined function that can be called from within an SQL statement.

    This function has two parameters.

    :param expr: pattern to look for.
    :type expr: str
    :param value: the string to search.
    :type value: list
    :returns: True if the expr exists within the value, false otherwise.
    :rtype: bool
    """
    return re.search(expr, value, re.MULTILINE) is not None



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Douglas S. Blank
# Copyright (C) 2019  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Age Stats Gramplet

This Gramplet shows distributions of age breakdowns of various types.
"""
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from collections import defaultdict

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gen.lib import Date, ChildRefType
from gramps.gui.widgets import Histogram
from gramps.gui.plug.quick import run_quick_report_by_name
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class AgeStatsGramplet(Gramplet):
    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

        self.max_age = 110
        self.max_mother_diff = 55
        self.max_father_diff = 70

    def db_changed(self):
        self.connect(self.dbstate.db, "person-add", self.update)
        self.connect(self.dbstate.db, "person-delete", self.update)
        self.connect(self.dbstate.db, "person-update", self.update)
        self.connect(self.dbstate.db, "event-update", self.update)

    def build_gui(self):
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.vbox.set_margin_start(6)
        self.vbox.set_margin_end(6)
        self.vbox.set_margin_top(6)
        self.vbox.set_margin_bottom(6)
        self.vbox.set_spacing(12)
        return self.vbox

    def build_options(self):
        from gramps.gen.plug.menu import NumberOption

        self.add_option(NumberOption(_("Max age"), self.max_age, 5, 150, 5))
        self.add_option(
            NumberOption(
                _("Max age of Mother at birth"), self.max_mother_diff, 5, 150, 5
            )
        )
        self.add_option(
            NumberOption(
                _("Max age of Father at birth"), self.max_father_diff, 5, 150, 5
            )
        )

    def save_options(self):
        self.max_age = int(self.get_option(_("Max age")).get_value())
        self.max_mother_diff = int(
            self.get_option(_("Max age of Mother at birth")).get_value()
        )
        self.max_father_diff = int(
            self.get_option(_("Max age of Father at birth")).get_value()
        )

    def on_load(self):
        if len(self.gui.data) == 3:
            self.max_age = int(self.gui.data[0])
            self.max_mother_diff = int(self.gui.data[1])
            self.max_father_diff = int(self.gui.data[2])

    def save_update_options(self, widget=None):
        self.max_age = int(self.get_option(_("Max age")).get_value())
        self.max_mother_diff = int(
            self.get_option(_("Max age of Mother at birth")).get_value()
        )
        self.max_father_diff = int(
            self.get_option(_("Max age of Father at birth")).get_value()
        )
        self.gui.data = [self.max_age, self.max_mother_diff, self.max_father_diff]
        self.update()

    def main(self):
        for widget in self.vbox.get_children():
            self.vbox.remove(widget)
        if not self.dbstate.is_open():
            return
        age_dict = defaultdict(int)
        mother_dict = defaultdict(int)
        father_dict = defaultdict(int)
        age_handles = defaultdict(list)
        mother_handles = defaultdict(list)
        father_handles = defaultdict(list)
        text = ""
        count = 0
        for person in self.dbstate.db.iter_people():
            if count % 300 == 0:
                yield True
            # if birth_date and death_date, compute age
            birth_date = self.get_date("BIRTH", person)
            death_date = self.get_date("DEATH", person)
            if birth_date:
                if death_date:
                    age = (death_date - birth_date).tuple()[0]
                    if age >= 0:
                        age_dict[age] += 1
                        age_handles[age].append(person.handle)

                # for each parent m/f:
                mother, father = self.get_birth_parents(person)
                if mother:
                    bdate = self.get_date("BIRTH", mother)
                    if bdate:
                        diff = (birth_date - bdate).tuple()[0]
                        if diff >= 0:
                            mother_dict[diff] += 1
                            mother_handles[diff].append(mother.handle)
                if father:
                    bdate = self.get_date("BIRTH", father)
                    if bdate:
                        diff = (birth_date - bdate).tuple()[0]
                        if diff >= 0:
                            father_dict[diff] += 1
                            father_handles[diff].append(father.handle)

            count += 1

        self.create_histogram(
            age_dict,
            age_handles,
            _("Lifespan Age Distribution"),
            _("Age"),
            5,
            self.max_age,
        )
        self.create_histogram(
            father_dict,
            father_handles,
            _("Father - Child Age Diff Distribution"),
            _("Diff"),
            5,
            self.max_father_diff,
        )
        self.create_histogram(
            mother_dict,
            mother_handles,
            _("Mother - Child Age Diff Distribution"),
            _("Diff"),
            5,
            self.max_mother_diff,
        )

    def get_date(self, event_type, person):
        """
        Find the birth or death date of a given person.
        """
        if event_type == "BIRTH":
            ref = person.get_birth_ref()
        else:
            ref = person.get_death_ref()
        if ref:
            event = self.dbstate.db.get_event_from_handle(ref.ref)
            date = event.get_date_object()
            if date.is_valid():
                return date
        return None

    def get_birth_parents(self, person):
        """
        Find the biological parents of a given person.
        """
        m_handle = None
        f_handle = None
        family_list = person.get_parent_family_handle_list()
        for family_handle in family_list:
            family = self.dbstate.db.get_family_from_handle(family_handle)
            if family:
                childrel = [
                    (ref.get_mother_relation(), ref.get_father_relation())
                    for ref in family.get_child_ref_list()
                    if ref.ref == person.handle
                ]
                if childrel[0][0] == ChildRefType.BIRTH:
                    m_handle = family.get_mother_handle()
                if childrel[0][1] == ChildRefType.BIRTH:
                    f_handle = family.get_father_handle()
        mother = None
        father = None
        if m_handle:
            mother = self.dbstate.db.get_person_from_handle(m_handle)
        if f_handle:
            father = self.dbstate.db.get_person_from_handle(f_handle)
        return mother, father

    def compute_stats(self, data):
        """
        Create a table of statistics based on a dictionary of data.
        """
        keys = sorted(data)
        count = sum(data.values())
        sumval = sum(k * data[k] for k in data)
        minval = min(keys)
        maxval = max(keys)
        median = 0
        average = 0
        if count > 0:
            current = 0
            for k in keys:
                if current + data[k] > count / 2:
                    median = k
                    break
                current += data[k]
            average = sumval / float(count)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        label = Gtk.Label(label=_("Statistics") + ":")
        label.set_halign(Gtk.Align.START)
        vbox.pack_start(label, False, False, 0)
        grid = Gtk.Grid()
        grid.set_margin_start(12)
        grid.set_column_spacing(12)
        rows = [
            [_("Total"), "%d" % count],
            [_("Minimum"), "%d" % minval],
            [_("Average"), glocale.format_string("%.1f", average)],
            [_("Median"), "%d" % median],
            [_("Maximum"), "%d" % maxval],
        ]
        for row, value in enumerate(rows):
            label1 = Gtk.Label(label=value[0] + ":")
            label1.set_halign(Gtk.Align.START)
            grid.attach(label1, 0, row, 1, 1)
            label2 = Gtk.Label(label=value[1])
            label2.set_halign(Gtk.Align.END)
            grid.attach(label2, 1, row, 1, 1)
        vbox.pack_start(grid, False, False, 0)
        vbox.show_all()

        return vbox

    def create_histogram(self, data, handles, title, column, interval, max_val):
        """
        Create a histogram based on a dictionary of data, like:
        data = {12: 4, 20: 6, 35: 13, 50: 5}
        where the key is the age, and the value stored is the count.
        """
        if len(data) == 0:
            return

        buckets = [0] * (int(max_val / interval) + 1)
        handle_data = defaultdict(list)
        for value, count in data.items():
            if value > max_val:
                buckets[int(max_val / interval)] += count
                handle_data[int(max_val / interval)].extend(handles[value])
            else:
                buckets[int(value / interval)] += count
                handle_data[int(value / interval)].extend(handles[value])

        labels = []
        for i in range(int(max_val / interval)):
            labels.append(
                "%d-%d"
                % (
                    i * interval,
                    (i + 1) * interval,
                )
            )
        labels.append("%d+" % ((i + 1) * interval,))

        hist = Histogram()
        hist.set_title(title)
        hist.set_bucket_axis(column)
        hist.set_value_axis("%")
        hist.set_values(buckets)
        hist.set_labels(labels)
        hist.set_tooltip(_("Double-click to see %d people"))
        hist.set_highlight([len(labels) - 1])
        hist.connect("clicked", self.on_bar_clicked, handle_data)
        hist.show()
        self.vbox.pack_start(hist, True, True, 0)

        stats = self.compute_stats(data)
        self.vbox.pack_start(stats, False, False, 0)

    def on_bar_clicked(self, _dummy, value, handle_data):
        """
        Called when a histogram bar is double-clicked.
        """
        run_quick_report_by_name(
            self.gui.dbstate,
            self.gui.uistate,
            "filterbyname",
            "list of people",
            handles=handle_data[value],
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2008       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The base option class for all other option classes.
"""

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from ...utils.callback import Callback


# -------------------------------------------------------------------------
#
# Option class
#
# -------------------------------------------------------------------------
class Option(Callback):
    """
    This class serves as a base class for all options. All Options must
    minimally provide the services provided by this class. Options are allowed
    to add additional functionality.
    """

    __signals__ = {"value-changed": None, "avail-changed": None}

    def __init__(self, label, value):
        """
        :param label: A friendly label to be applied to this option.
            Example: "Exclude living people"
        :type label: string
        :param value: An initial value for this option.
            Example: True
        :type value: The type will depend on the type of option.
        :return: nothing
        """
        Callback.__init__(self)
        self.__value = value
        self.__label = label
        self.__help_str = ""
        self.__available = True

    def get_label(self):
        """
        Get the friendly label for this option.

        :return: string
        """
        return self.__label

    def set_label(self, label):
        """
        Set the friendly label for this option.

        :param label: A friendly label to be applied to this option.
            Example: "Exclude living people"
        :type label: string
        :return: nothing
        """
        self.__label = label

    def get_value(self):
        """
        Get the value of this option.

        :return: The option value.
        """
        return self.__value

    def set_value(self, value):
        """
        Set the value of this option.

        :param value: A value for this option.
            Example: True
        :type value: The type will depend on the type of option.
        :return: nothing
        """
        self.__value = value
        self.emit("value-changed")

    def get_help(self):
        """
        Get the help information for this option.

        :return: A string that provides additional help beyond the label.
        """
        return self.__help_str

    def set_help(self, help_text):
        """
        Set the help information for this option.

        :param help: A string that provides additional help beyond the label.
            Example: "Whether to include or exclude people who are calculated
            to be alive at the time of the generation of this report"
        :type value: string
        :return: nothing
        """
        self.__help_str = help_text

    def set_available(self, avail):
        """
        Set the availability of this option.

        :param avail: An indicator of whether this option is currently
                      available. True indicates that the option is available.
                      False indicates that the option is not available.
        :type avail: Bool
        :return: nothing
        """
        if avail != self.__available:
            self.__available = avail
            self.emit("avail-changed")

    def get_available(self):
        """
        Get the availability of this option.

        :return: A Bool indicating the availablity of this option. True
                 indicates that the option is available. False indicates that
                 the option is not available.
        """
        return self.__available



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".filter")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
# we need global variableCustomFilters, so we need to query gramps.gen.filters
# when we need this variable, not import it at the start!
import gramps.gen.filters
from . import Rule
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Any, Union
from ...db import Database


# -------------------------------------------------------------------------
#
# MatchesFilter
#
# -------------------------------------------------------------------------
class MatchesFilterBase(Rule):
    """
    Rule that checks against another filter.

    This is a base rule for subclassing by specific objects.
    Subclasses need to define the namespace class attribute.

    """

    labels = [_("Filter name:")]
    name = "Objects matching the <filter>"
    description = "Matches objects matched by the specified filter name"
    category = _("General filters")
    namespace: Union[str, None] = None

    def prepare(self, db: Database, user):
        if gramps.gen.filters.CustomFilters:
            filters = gramps.gen.filters.CustomFilters.get_filters_dict(self.namespace)
            if self.list[0] in filters:
                filt = filters[self.list[0]]
                for rule in filt.flist:
                    rule.requestprepare(db, user)
            else:
                LOG.warning(
                    _("Can't find filter %s in the defined custom filters")
                    % self.list[0]
                )
        else:
            LOG.warning(
                _("Can't find filter %s in the defined custom filters") % self.list[0]
            )

    def reset(self):
        if gramps.gen.filters.CustomFilters:
            filters = gramps.gen.filters.CustomFilters.get_filters_dict(self.namespace)
            if self.list[0] in filters:
                filt = filters[self.list[0]]
                for rule in filt.flist:
                    rule.requestreset()

    def apply_to_one(self, db: Database, obj: Any) -> bool:
        if gramps.gen.filters.CustomFilters:
            filters = gramps.gen.filters.CustomFilters.get_filters_dict(self.namespace)
            if self.list[0] in filters:
                filt = filters[self.list[0]]
                return filt.apply_to_one(db, obj)
        return False

    def find_filter(self):
        """
        Return the selected filter or None.
        """
        if gramps.gen.filters.CustomFilters:
            filters = gramps.gen.filters.CustomFilters.get_filters_dict(self.namespace)
            if self.list[0] in filters:
                return filters[self.list[0]]
        return None



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2025      Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Unittest for DataDict"""

# See also tests in ./serialize_test.py

import unittest

from gramps.gen.lib.json_utils import DataDict, DataList, object_to_data, data_to_object
from gramps.gen.lib import (
    Person,
    Family,
)


class DataDictTest(unittest.TestCase):
    def test_person(self):
        p = Person()
        d = DataDict(p)
        # Get the object from the dict:
        self.assertEqual(id(p), id(data_to_object(d)))

    def test_person_to_data(self):
        p = Person()
        d = object_to_data(p)
        self.assertFalse(hasattr(d, "_object"))

    def test_family(self):
        p = Family()
        d = DataDict(p)
        # Get the object from the dict:
        self.assertEqual(id(p), id(data_to_object(d)))

    def test_family_to_data(self):
        p = Family()
        d = object_to_data(p)
        self.assertFalse(hasattr(d, "_object"))


class DataListTest(unittest.TestCase):
    def test_empty_1(self):
        dl = DataList()
        self.assertIsInstance(dl, DataList)
        self.assertEqual(len(dl), 0)

    def test_empty_2(self):
        dl = DataList([])
        self.assertIsInstance(dl, DataList)
        self.assertEqual(len(dl), 0)

    def test_value_1(self):
        dl = DataList([42])
        self.assertIsInstance(dl, DataList)
        self.assertIsInstance(dl[0], int)
        self.assertEqual(dl[0], 42)

    def test_value_exception(self):
        with self.assertRaises(TypeError):
            dl = DataList(42)

    def test_access_exception(self):
        dl = DataList([1, 2, 3])
        self.assertIsInstance(dl, DataList)
        self.assertEqual(dl[0], 1)
        self.assertEqual(dl[1], 2)
        self.assertEqual(dl[2], 3)
        with self.assertRaises(IndexError):
            dl[3]

    def test_nested(self):
        dl = DataList([[42]])
        self.assertIsInstance(dl, DataList)
        self.assertIsInstance(dl[0], DataList)
        self.assertEqual(dl[0][0], 42)

    def test_data(self):
        p = Person()
        p_data = object_to_data(p)
        dl = DataList([p_data])
        self.assertIsInstance(dl, DataList)
        self.assertIsInstance(dl[0], DataDict)
        self.assertEqual(dl[0], p_data)
        self.assertEqual(dl[0].gender, 2)

    def test_append_list(self):
        dl = DataList([])
        dl.append([])
        self.assertIsInstance(dl, DataList)
        self.assertEqual(dl[0], [])
        self.assertIsInstance(dl[0], DataList)

    def test_append_value(self):
        dl = DataList([])
        dl.append(42)
        self.assertIsInstance(dl, DataList)
        self.assertEqual(dl[0], 42)
        self.assertIsInstance(dl[0], int)

    def test_combined_list(self):
        dl = DataList([])
        self.assertIsInstance(dl + [], DataList)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
# Copyright (C) 2013       Zissis Papadopoulos <zissis@mail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Greek-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Greek parser class
#
# -------------------------------------------------------------------------
class DateParserEL(DateParser):
    """
    Convert a text string into a Date object. If the date cannot be
    converted, the text string is assigned.
    """

    # modifiers before the date
    modifier_to_int = {
        "προ του": Date.MOD_BEFORE,
        "πριν το": Date.MOD_BEFORE,
        "πριν από τις": Date.MOD_BEFORE,
        "πριν από την": Date.MOD_BEFORE,
        "πριν από το": Date.MOD_BEFORE,
        "πριν από τον": Date.MOD_BEFORE,
        "προ": Date.MOD_BEFORE,
        "πρ.": Date.MOD_BEFORE,
        "μετά το": Date.MOD_AFTER,
        "μετά από τις": Date.MOD_AFTER,
        "μετά από την": Date.MOD_AFTER,
        "μετά από το": Date.MOD_AFTER,
        "μετά από τον": Date.MOD_AFTER,
        "μετά": Date.MOD_AFTER,
        "μετ.": Date.MOD_AFTER,
        "γύρω στο": Date.MOD_ABOUT,
        "γύρω στον": Date.MOD_ABOUT,
        "γύρω στις": Date.MOD_ABOUT,
        "περίπου το": Date.MOD_ABOUT,
        "περ.": Date.MOD_ABOUT,
        "γυρ.": Date.MOD_ABOUT,
        "~": Date.MOD_ABOUT,
        "from": Date.MOD_FROM,
        "to": Date.MOD_TO,
    }

    bce = ["π.Χ.", "π.Κ.Χ.", "π.Κ.Ε.", "π.Χ"]

    calendar_to_int = {
        "γρηγοριανό": Date.CAL_GREGORIAN,
        "γ": Date.CAL_GREGORIAN,
        "ιουλιανό": Date.CAL_JULIAN,
        "ι": Date.CAL_JULIAN,
        "εβραϊκό": Date.CAL_HEBREW,
        "ε": Date.CAL_HEBREW,
        "ισλαμικό": Date.CAL_ISLAMIC,
        "ισλ": Date.CAL_ISLAMIC,
        "γαλλικό": Date.CAL_FRENCH,
        "γαλλικής δημοκρατίας": Date.CAL_FRENCH,
        "γ": Date.CAL_FRENCH,
        "περσικό": Date.CAL_PERSIAN,
        "π": Date.CAL_PERSIAN,
        "σουηδικό": Date.CAL_SWEDISH,
        "σ": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "κατʼ εκτίμηση": Date.QUAL_ESTIMATED,
        "εκτιμώμενη": Date.QUAL_ESTIMATED,
        "εκτ.": Date.QUAL_ESTIMATED,
        "εκτ": Date.QUAL_ESTIMATED,
        "υπολογ": Date.QUAL_CALCULATED,
        "υπολογ.": Date.QUAL_CALCULATED,
        "υπολογισμένη": Date.QUAL_CALCULATED,
        "με υπολογισμό": Date.QUAL_CALCULATED,
    }

    def init_strings(self):
        """
        This method compiles regular expression strings for matching dates.
        """
        DateParser.init_strings(self)
        _span_1 = ["από"]
        _span_2 = ["έως"]
        _range_1 = ["μετ", r"μετ\.", "μεταξύ"]
        _range_2 = ["και"]
        self._span = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )


# -------------------------------------------------------------------------
#
# Greek display
#
# -------------------------------------------------------------------------
class DateDisplayEL(DateDisplay):
    """
    Greek language date display class.
    """

    # this is used to display the 12 gregorian months
    long_months = (
        "",
        "Ιανουάριος",
        "Φεβρουάριος",
        "Μάρτιος",
        "Απρίλιος",
        "Μάιος",
        "Ιούνιος",
        "Ιούλιος",
        "Αύγουστος",
        "Σεπτέμβριος",
        "Οκτώβριος",
        "Νοέμβριος",
        "Δεκέμβριος",
    )

    short_months = (
        "",
        "Ιαν",
        "Φεβ",
        "Μαρ",
        "Απρ",
        "Μάι",
        "Ιουν",
        "Ιουλ",
        "Αύγ",
        "Σεπ",
        "Οκτ",
        "Νοε",
        "Δεκ",
    )

    _mod_str = ("", "προ του ", "μετά το ", "γύρω στο ", "", "", "")

    _qual_str = ("", "εκτιμώμενη ", "υπολογισμένη ")

    _bce_str = "%s π.Χ."

    formats = (
        "ΕΕΕΕ-ΜΜ-ΗΗ (ISO)",
        "ΗΗ-ΜΜ-ΕΕΕΕ",
        "ΗΗ/ΜΜ/ΕΕΕΕ",
        "ΗΗ Μήνας ΕΕΕΕ",
        "ΗΗ Μήν ΕΕΕΕ",
    )
    # this definition must agree with its "_display_gregorian" method

    def _display_gregorian(self, date_val, **kwargs):
        """
        display gregorian calendar date in different format
        """
        # this must agree with its locale-specific "formats" definition
        year = self._slash_year(date_val[2], date_val[3])
        if self.format == 0:
            return self.display_iso(date_val)
        elif self.format == 1:
            # day-month_number-year
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = year
                else:
                    value = "%s-%s" % (date_val[1], year)
            else:
                value = "%d-%s-%s" % (date_val[0], date_val[1], year)
        elif self.format == 2:
            # day/month_number/year
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = year
                else:
                    value = "%s/%s" % (date_val[1], year)
            else:
                value = "%d/%s/%s" % (date_val[0], date_val[1], year)
        elif self.format == 3:
            # day month_name year
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = year
                else:
                    value = "%s %s" % (self.long_months[date_val[1]], year)
            else:
                value = "%d %s %s" % (date_val[0], self.long_months[date_val[1]], year)
        else:
            # day month_abbreviation year
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = year
                else:
                    value = "%s %s" % (self.short_months[date_val[1]], year)
            else:
                value = "%d %s %s" % (date_val[0], self.short_months[date_val[1]], year)
        if date_val[2] < 0:
            return self._bce_str % value
        else:
            return value

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "από", d1, "έως", d2, scal)
        elif mod == Date.MOD_RANGE:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "μεταξύ", d1, "και", d2, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            return "%s%s%s%s" % (qual_str, self._mod_str[mod], text, scal)


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("el_GR", "el_CY", "el", "Greek", "greek", ("%d/%m/%Y",)),
    DateParserEL,
    DateDisplayEL,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide Error objects
"""


class FilterError(Exception):
    """Error used to report Filter errors"""

    def __init__(self, value, value2=""):
        Exception.__init__(self)
        self.value = value
        self.value2 = value2

    def __str__(self):
        "Return string representation"
        return self.value

    def messages(self):
        "Return the messages"
        return (self.value, self.value2)


class DateError(Exception):
    """Error used to report Date errors

    Might have a .date attribute holding an invalid Date object
    that triggered the error.
    """

    def __init__(self, value=""):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        "Return string representation"
        return self.value


class DatabaseError(Exception):
    """Error used to report database errors"""

    def __init__(self, value=""):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        "Return string representation"
        return self.value


class ReportError(Exception):
    """Error used to report Report errors."""

    def __init__(self, value, value2=""):
        Exception.__init__(self)
        self.value = value
        self.value2 = value2

    def __str__(self):
        "Return string representation"
        return self.value

    def messages(self):
        "Return the messages"
        return (self.value, self.value2)


class GedcomError(Exception):
    """Error used to report GEDCOM errors"""

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        "Return string representation"
        return self.value


class GrampsImportError(Exception):
    """Error used to report mistakes during import of files into Gramps"""

    def __init__(self, value, value2=""):
        Exception.__init__(self)
        self.value = value
        self.value2 = value2

    def __str__(self):
        "Return string representation"
        return self.value

    def messages(self):
        "Return the messages"
        return (self.value, self.value2)


class PluginError(Exception):
    """Error used to report plugin errors"""

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        "Return string representation"
        return self.value


class HandleError(Exception):
    """Error used to report wrong database handle errors"""

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        "Return string representation"
        return self.value


class WindowActiveError(Exception):
    """Error used to report that the request window is already displayed."""

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        "Return string representation"
        return self.value


class UnavailableError(Exception):
    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        "Return string representation"
        return self.value


class MaskError(Exception):
    pass


class ValidationError(Exception):
    pass


class DbError(Exception):
    """Error used to report BerkeleyDB errors."""

    def __init__(self, value):
        Exception.__init__(self)
        try:
            (errnum, errmsg) = value
            self.value = errmsg
        except:
            self.value = value

    def __str__(self):
        "Return string representation"
        return self.value


class MergeError(Exception):
    """Error used to report merge errors"""

    def __init__(self, value=""):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        "Return string representation"
        return self.value



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Rule that checks for a note of a particular type.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale
from ....lib.notetype import NoteType
from .. import Rule

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Note
from ....db import Database


# -------------------------------------------------------------------------
#
# HasType
#
# -------------------------------------------------------------------------
class HasType(Rule):
    """
    Rule that checks for a note of a particular type.
    """

    labels = [_("Note type:")]
    name = _("Notes with the particular type")
    description = _("Matches notes with the particular type ")
    category = _("General filters")

    def __init__(self, arg, use_regex=False, use_case=False):
        super().__init__(arg, use_regex, use_case)
        self.note_type = None

    def prepare(self, db: Database, user):
        """
        Prepare the rule. Things we only want to do once.
        """
        if self.list[0]:
            self.note_type = NoteType()
            self.note_type.set_from_xml_str(self.list[0])

    def apply_to_one(self, _db: Database, obj: Note) -> bool:
        """
        Apply the rule. Return True on a match.
        """
        if self.note_type:
            return obj.type == self.note_type
        return False



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2009       Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2014       Vlada Perić <vlada.peric@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
from __future__ import annotations
from xml.parsers import expat
import datetime
import math
import os

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import PLUGINS_DIR, USER_PLUGINS, DATA_DIR
from gramps.gen.lib.gcalendar import gregorian_ymd, hebrew_sdn


# ------------------------------------------------------------------------
#
# Support functions
#
# ------------------------------------------------------------------------
def g2iso(dow):
    """Converst Gramps day of week to ISO day of week"""
    # Gramps: SUN = 1
    # ISO: MON = 1
    return (dow + 5) % 7 + 1


def easter(year):
    """
    Computes the year/month/day of easter. Based on work by
    J.-M. Oudin (1940) and is reprinted in the "Explanatory Supplement
    to the Astronomical Almanac", ed. P. K.  Seidelmann (1992).  Note:
    Ash Wednesday is 46 days before Easter Sunday.
    """
    c = year // 100
    n = year - 19 * (year // 19)
    k = (c - 17) // 25
    i = c - c // 4 - (c - k) // 3 + 19 * n + 15
    i = i - 30 * (i // 30)
    i = i - (i // 28) * (1 - (i // 28) * (29 // (i + 1)) * ((21 - n) // 11))
    j = year + year // 4 + i + 2 - c + c // 4
    j = j - 7 * (j // 7)
    l = i - j
    month = 3 + (l + 40) // 44
    day = l + 28 - 31 * (month // 4)
    return "%d/%d/%d" % (year, month, day)


def julian_easter(year):
    """
    Computes the year/month/day of Eastern Orthodox Easter, given in the
    Gregorian calendar. Implements the Jean Meeus algorithm. Valid: 1900-2099.
    """
    a = year % 4
    b = year % 7
    c = year % 19
    d = (19 * c + 15) % 30
    e = (2 * a + 4 * b - d + 34) % 7
    month = int(math.floor((d + e + 114) / 31))
    day = ((d + e + 114) % 31) + 1

    # produced date was in the Julian calendar, add 13 days to it
    day = day + 13

    if month == 3 and day > 31:
        day = day - 31
        month = 4
    elif month == 4 and day > 30:
        day = day - 30
        month = 5

    return "%d/%d/%d" % (year, month, day)


def passover(year):
    """
    Returns the date of Passover in a given Gregorian year.
    """
    heb_year = year + 3760
    heb = hebrew_sdn(heb_year, 8, 15)  # Passover, 15 Nissan

    return "%d/%d/%d" % gregorian_ymd(heb)


def hanuka(year):
    """
    Returns the date of first day of Hanuka in a given Gregorian year.
    We can't use passover as an offset, because the year length changes.
    The hebrew year have 6 possible lengths.
    """
    heb_year = (
        year + 3761
    )  # Not 3760, because Hanuka is in Nov/Dec of the previous year
    heb = hebrew_sdn(heb_year, 3, 25)  # Hanuka, 25 Kislev

    return "%d/%d/%d" % gregorian_ymd(heb)


def dst(year, area="us"):
    """
    Return Daylight Saving Time start/stop in a given area ("us", "eu").
    US calculation valid 1976-2099; EU 1996-2099
    """
    if area == "us":
        if year > 2006:
            start = "%d/%d/%d" % (
                year,
                3,
                14 - (math.floor(1 + year * 5 / 4) % 7),
            )  # March
            stop = "%d/%d/%d" % (
                year,
                11,
                7 - (math.floor(1 + year * 5 / 4) % 7),
            )  # November
        else:
            start = "%d/%d/%d" % (
                year,
                4,
                (2 + 6 * year - math.floor(year / 4)) % 7 + 1,
            )  # April
            stop = "%d/%d/%d" % (
                year,
                10,
                (31 - (math.floor(year * 5 / 4) + 1) % 7),
            )  # October
    elif area == "eu":
        start = "%d/%d/%d" % (
            year,
            3,
            (31 - (math.floor(year * 5 / 4) + 4) % 7),
        )  # March
        stop = "%d/%d/%d" % (year, 10, (31 - (math.floor(year * 5 / 4) + 1) % 7))  # Oct
    return (start, stop)


def dow(y, m, d):
    """Return the ISO day of week for the given year, month and day."""
    return datetime.date(y, m, d).isoweekday()


def cmp(a, b):
    """
    Replacement for older Python's cmp.
    """
    return (a > b) - (a < b)


# ------------------------------------------------------------------------
#
# HolidayTable
#
# ------------------------------------------------------------------------
class HolidayTable:
    """
    HolidayTable is a class which provides holidays for various
    countries and years.
    """

    __holiday_files: list[str] = []
    __countries: list[str] = []

    def __init__(self):
        """
        Find the holiday files and load the countries if it has not already
        been done.
        """
        if not HolidayTable.__holiday_files:
            self.__find_holiday_files()

        if not HolidayTable.__countries:
            self.__build_country_list()

        # Initialize the holiday table to be empty
        self.__holidays = {}
        self.__init_table()

    def __find_holiday_files(self):
        """
        Looks in multiple places for holidays.xml files
        It will search for the file in user's plugin directories first,
        then it will search in program's plugins directories.
        """

        holiday_file = "holidays.xml"

        # Look for holiday files in the user plugins directory and all
        # subdirectories.
        for dirpath, dirnames, filenames in os.walk(USER_PLUGINS):
            holiday_full_path = os.path.join(dirpath, holiday_file)
            if os.path.exists(holiday_full_path):
                HolidayTable.__holiday_files.append(holiday_full_path)

        # Look for holiday files in the installation data directory and all
        # subdirectories.
        for dirpath, dirnames, filenames in os.walk(DATA_DIR):
            holiday_full_path = os.path.join(dirpath, holiday_file)
            if os.path.exists(holiday_full_path):
                HolidayTable.__holiday_files.append(holiday_full_path)

    def __build_country_list(self):
        """Generate the list of countries that have holiday information."""
        for holiday_file_path in HolidayTable.__holiday_files:
            parser = _Xml2Obj()
            root_element = parser.parse(holiday_file_path)
            for country_element in root_element.get_children():
                if country_element.get_name() == "country":
                    country_name = country_element.get_attribute("name")
                    if country_name not in HolidayTable.__countries:
                        HolidayTable.__countries.append(_(country_name))

    def __init_table(self):
        """Initialize the holiday table structure."""
        for month in range(1, 13):
            self.__holidays[month] = {}
            for day in range(1, 32):
                self.__holidays[month][day] = []

    def get_countries(self):
        """
        Get all the country names that holidays are available for.

        @return: nothing
        """
        return HolidayTable.__countries

    def load_holidays(self, year, country):
        """
        Load the holiday table for the specified year and country.
        This must be called before get_holidays().

        @param year: The year for which the holidays should be loaded.
            Example: 2010
        @type year: int
        @param country: The country for which the holidays should be loaded.
            Example: "United States"
        @type country: str
        @return: nothing
        """
        self.__init_table()
        for holiday_file_path in HolidayTable.__holiday_files:
            parser = _Xml2Obj()
            element = parser.parse(holiday_file_path)
            calendar = _Holidays(element, country)
            date = datetime.date(year, 1, 1)
            while date.year == year:
                holidays = calendar.check_date(date)
                for text in holidays:
                    self.__holidays[date.month][date.day].append(_(text))
                date = date.fromordinal(date.toordinal() + 1)

    def get_holidays(self, month, day):
        """
        Get the holidays for the given day of the year.

        @param month: The month for the requested holidays.
            Example: 1
        @type month: int
        @param month: The day for the requested holidays.
            Example: 1
        @type month: int
        @return: An array of strings with holiday names.
        @return type: [str]
        """
        return self.__holidays[month][day]


# ------------------------------------------------------------------------
#
# _Element
#
# ------------------------------------------------------------------------
class _Element:
    """A parsed XML element"""

    def __init__(self, name, attributes):
        "Element constructor"
        # The element's tag name
        self.__name = name
        # The element's attribute dictionary
        self.__attributes = attributes
        # The element's child element list (sequence)
        self.__children = []

    def add_child(self, element):
        "Add a reference to a child element"
        self.__children.append(element)

    def get_attribute(self, key):
        "Get an attribute value"
        return self.__attributes.get(key)

    def get_attributes(self):
        "Get all the attributes"
        return self.__attributes

    def get_name(self):
        """Get the name of this element."""
        return self.__name

    def get_children(self):
        """Get the children elements for this element."""
        return self.__children


# ------------------------------------------------------------------------
#
# _Xml2Obj
#
# ------------------------------------------------------------------------
class _Xml2Obj:
    """XML to Object"""

    def __init__(self):
        self.root = None
        self.nodeStack = []

    def start_element(self, name, attributes):
        "SAX start element even handler"
        # Instantiate an Element object
        element = _Element(name, attributes)
        # Push element onto the stack and make it a child of parent
        if len(self.nodeStack) > 0:
            parent = self.nodeStack[-1]
            parent.add_child(element)
        else:
            self.root = element
        self.nodeStack.append(element)

    def end_element(self, name):
        "SAX end element event handler"
        self.nodeStack = self.nodeStack[:-1]

    def parse(self, filename):
        "Create a SAX parser and parse filename"
        parser = expat.ParserCreate()
        # SAX event handlers
        parser.StartElementHandler = self.start_element
        parser.EndElementHandler = self.end_element
        # Parse the XML File
        with open(filename, "rb") as xml_file:
            parser.ParseFile(xml_file)
        return self.root


# ------------------------------------------------------------------------
#
# _Holidays
#
# ------------------------------------------------------------------------
class _Holidays:
    """Class used to read XML holidays to add to calendar."""

    MONTHS = [
        "jan",
        "feb",
        "mar",
        "apr",
        "may",
        "jun",
        "jul",
        "aug",
        "sep",
        "oct",
        "nov",
        "dec",
    ]
    DAYS = ["mon", "tue", "wed", "thu", "fri", "sat", "sun"]
    WORKDAY = list(range(5))  # indexes into above
    WEEKEND = (5, 6)  # indexes into above

    def __init__(self, elements, country="US"):
        self.debug = 0
        self.elements = elements
        self.country = country
        self.dates = []
        self.initialize()

    def set_country(self, country):
        """Set the contry of holidays to read"""
        self.country = country
        self.dates = []
        self.initialize()

    def initialize(self):
        """Parse the holiday date XML items"""
        for country_set in self.elements.get_children():
            if (
                country_set.get_name() == "country"
                and _(country_set.get_attribute("name")) == self.country
            ):
                for date in country_set.get_children():
                    if date.get_name() == "date":
                        data = {
                            "value": "",
                            "name": "",
                            "offset": "",
                            "type": "",
                            "if": "",
                        }  # defaults
                        for attr in date.get_attributes():
                            data[attr] = date.get_attribute(attr)
                        self.dates.append(data)

    def get_daynames(self, year, month, dayname):
        """Get the items for a particular year/month and day of week"""
        if self.debug:
            print("%s's in %d %d..." % (dayname, month, year))

        retval = [0]
        day_of_week = self.DAYS.index(dayname)
        for day in range(1, 32):
            try:
                date = datetime.date(year, month, day)
            except ValueError:
                continue
            if date.weekday() == day_of_week:
                retval.append(day)

        if self.debug:
            print("day_of_week=", day_of_week, "days=", retval)

        return retval

    def check_date(self, date):
        """Return items that match rules"""
        retval = []
        for rule in self.dates:
            if self.debug:
                print("Checking ", rule["name"], "...")

            offset = 0
            if rule["offset"] != "":
                if rule["offset"].isdigit():
                    offset = int(rule["offset"])
                elif rule["offset"][0] in ["-", "+"] and rule["offset"][1:].isdigit():
                    offset = int(rule["offset"])
                else:
                    # must be a dayname or "workday"
                    offset = rule["offset"]

            if rule["value"].startswith(">"):
                # eval exp -> year/num[/day[/month]]
                y, m, d = date.year, date.month, date.day
                rule["value"] = eval(rule["value"][1:])

            if self.debug:
                print("rule['value']:", rule["value"])
            if (
                rule["value"].count("/") == 3
            ):  # year/num/day/month, "3rd wednesday in april"
                y, num, dayname, mon = rule["value"].split("/")
                if y == "*":
                    y = date.year
                else:
                    y = int(y)
                if mon.isdigit():
                    m = int(mon)
                elif mon == "*":
                    m = date.month
                elif mon in self.MONTHS:
                    m = self.MONTHS.index(mon) + 1
                dates_of_dayname = self.get_daynames(y, m, dayname)

                if self.debug:
                    print("num =", num)

                d = dates_of_dayname[int(num)]

            elif rule["value"].count("/") == 2:  # year/month/day
                y, m, d = rule["value"].split("/")
                if y == "*":
                    y = date.year
                else:
                    y = int(y)
                if m == "*":
                    m = date.month
                elif m in self.MONTHS:
                    m = self.MONTHS.index(m) + 1
                else:
                    m = int(m)
                if d == "*":
                    d = date.day
                else:
                    d = int(d)
            ndate = datetime.date(y, m, d)

            if self.debug:
                print(ndate, offset, type(offset))

            if isinstance(offset, int):
                if offset != 0:
                    ndate = ndate.fromordinal(ndate.toordinal() + offset)
            elif isinstance(offset, str):
                direction = 1
                if offset[0] == "-":
                    direction = -1
                    offset = offset[1:]
                elif offset[0] == "+":
                    direction = 1
                    offset = offset[1:]

                if offset == "workday":
                    # next workday you come to, including this one
                    day_of_week = self.WORKDAY
                    ordinal = ndate.toordinal()
                    while ndate.fromordinal(ordinal).weekday() not in day_of_week:
                        ordinal += direction
                    ndate = ndate.fromordinal(ordinal)
                elif offset == "weekend":
                    # next weekend you come to, including this one
                    day_of_week = self.WEEKEND
                    ordinal = ndate.toordinal()
                    while ndate.fromordinal(ordinal).weekday() not in day_of_week:
                        ordinal += direction
                    ndate = ndate.fromordinal(ordinal)
                elif offset in self.DAYS:
                    # next tuesday you come to, including this one
                    day_of_week = self.DAYS.index(offset)
                    ordinal = ndate.toordinal()
                    while ndate.fromordinal(ordinal).weekday() != day_of_week:
                        ordinal += direction
                    ndate = ndate.fromordinal(ordinal)

            if self.debug:
                print("ndate:", ndate, "date:", date)

            if ndate == date:
                if rule["if"] != "":
                    if not eval(rule["if"]):
                        continue
                retval.append(rule["name"])

        return retval



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Any
from ....db import Database


# -------------------------------------------------------------------------
#
# HasNameOf
#
# -------------------------------------------------------------------------
class SearchName(Rule):
    """Rule that checks for full or partial name matches"""

    labels = [_("Substring:")]
    name = _("People matching the <name>")
    description = _("Matches people with a specified (partial) name")
    category = _("General filters")

    def apply_to_one(self, db: Database, person: Any) -> bool:
        src = self.list[0].upper()
        if not src:
            return False

        for name in [person.primary_name] + person.alternate_names:
            for field in [
                name.first_name,
                name.get_surname(),
                name.suffix,
                name.title,
                name.nick,
                name.famnick,
                name.call,
            ]:
                if src and field.upper().find(src) != -1:
                    return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Configuration based utilities
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib import Researcher
from ..config import config


# -------------------------------------------------------------------------
#
# Configuration based utilities
#
# -------------------------------------------------------------------------
def get_researcher():
    """
    Return a new database owner with the default values from the config file.
    """
    name = config.get("researcher.researcher-name")
    address = config.get("researcher.researcher-addr")
    locality = config.get("researcher.researcher-locality")
    city = config.get("researcher.researcher-city")
    state = config.get("researcher.researcher-state")
    country = config.get("researcher.researcher-country")
    post_code = config.get("researcher.researcher-postal")
    phone = config.get("researcher.researcher-phone")
    email = config.get("researcher.researcher-email")

    owner = Researcher()
    owner.set_name(name)
    owner.set_address(address)
    owner.set_locality(locality)
    owner.set_city(city)
    owner.set_state(state)
    owner.set_country(country)
    owner.set_postal_code(post_code)
    owner.set_phone(phone)
    owner.set_email(email)

    return owner



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ._matchesfilter import MatchesFilter


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Set
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# IsChildOfFilterMatch
#
# -------------------------------------------------------------------------
class IsChildOfFilterMatch(Rule):
    """Rule that checks for a person that is a child
    of someone matched by a filter"""

    labels = [_("Filter name:")]
    name = _("Children of <filter> match")
    category = _("Family filters")
    description = _("Matches children of anybody matched by a filter")

    def prepare(self, db: Database, user):
        self.db = db
        self.selected_handles: Set[str] = set()
        self.filt = MatchesFilter(self.list)
        self.filt.requestprepare(db, user)
        if user:
            user.begin_progress(
                self.category,
                _("Retrieving all sub-filter matches"),
                db.get_number_of_people(),
            )
        for person in db.iter_people():
            if user:
                user.step_progress()
            if self.filt.apply_to_one(db, person):
                self.init_list(person)
        if user:
            user.end_progress()

    def reset(self):
        self.filt.requestreset()
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles

    def init_list(self, person: Person):
        if not person:
            return
        for fam_id in person.family_list:
            fam = self.db.get_family_from_handle(fam_id)
            if fam:
                self.selected_handles.update(
                    child_ref.ref for child_ref in fam.child_ref_list
                )
